#!/usr/bin/perl

#
# $Id: sc_ldap2dnsdhcpd.pl 10073 2014-11-07 13:32:32Z HZuchel $
#

=encoding utf8

=head1 NAME

C<sc_ldap2dnsdhcpd.pl>



=head1 SYNOPSIS

=begin comment

  kurze Beschreibung, was das Programm macht
  und aufruf Parameter

=end comment


program for creating ...

    $PROGRAM
        --help




=head1 DESCRIPTION

=head2 Allg. Beschreibung und Aufgaben

=head2 Funktionsweise / Design / Architektur

=head2 Sicherheitskonzept

=head2 Einordnung / Abhaengigkeiten / Voraussetzungen

=over 4

=item Einordnung

=item Abhängigkeiten

=item Voraussetzungen

=back



=head2 Interfaces

=begin comment

    Kommandozeilenoptionen werden unter SYNOPSIS
    aufgeführt und in der usage function beschrieben

=end comment

=over 4

=item Programm

=item Benutzer

=back



=head2 Status (Fehler)



=head2 Installation und sonstige Hinweise



=head1 AUTHOR

    smartclient@dass-it.de




=head1 SEE ALSO



=head1 TODO



=cut

#
# used modules
#
use strict;
use warnings;
use Data::Dumper;
use Log::Log4perl;
use File::Basename;
use File::Copy;
use File::Compare;
use File::Temp qw( tempfile tempdir );
use IO::File;
use POSIX qw(tmpnam);
use Time::localtime;
use Time::HiRes qw( gettimeofday tv_interval );
use URI::Escape;
use SmartClient::Config;
use SmartClient::Constant;
## don't import all methods, because compare is also defined in File::Compare
use SmartClient::Util
  qw( stringArrayCompare getUndefinedHashKeys checkDir checkFileMod checkSymLink checkRemoveSymLink getUrlPort );
use SmartClient::LdapUtil;
use SmartClient::SharedFunction;
use SmartClient::Leases2ldap;
use SmartClient::RpcXmlUtil;
use SmartClient::LDAP;
use SmartClient::ScCache;

use constant DEFAULT_SPACER => q{    };
use constant INDENT_SPACER  => q{  };
use constant ONE_SPACE      => q{ };
use constant EMPTY_STR      => q{};

#
# declarations
#

=head1 Constants

=head1 Global Variables

=head2 $global_sc_ldap

SmartClient::LDAP object

=cut

my $global_sc_ldap;

=head2 $global_sc_cache

SmartClient::ScCache object

This is necessary for giving functions called via
SmartClient::SharedFunction access to the cache

=cut

my $global_sc_cache;

=head2 $global_leases2ldap

SmartClient::Leases2ldap object

=cut

my $global_leases2ldap;

=head2 $global_current_dhcp_subdir

To secure the creation of dhcp config files, two subdirectories will be used in turn.
They will be located under C<LDAP2DHCP_DHCPCONFPATH>.

This variable is an integer with values either C<0> or C<1>.

=cut

my $global_current_dhcp_subdir = 0;

=head2 $global_create_config_trigger

global variable to trigger the call of refreshEntryCache
in the global SmartClient::ScCache ($global_sc_cache)
on a secondary HA system

=cut

my $global_create_config_trigger = 0;

=head1 Functions

=head2 isPrimary

	checks configuration,
	if this is the primary HA system or not

=over 4

=item Parameter

    href SmartClient::Config object

=item Return

	boolean

=back

=cut

sub isPrimary
{
    my ($config) = @_;
    return $config->get( 'LDAP2DNSDHCPD_HA_MODE' ) eq 'ha_mode_primary';
}

=head2 readZoneTemplate

Read content of zone template header file

=over 4

=item Parameter

 - file name including path of zone template header file 

=item Return

 Array of lines read from the template file or undef in case
 of error.

=back

=cut

sub readZoneTemplate
{
    my ($template_filename) = @_;

    my @template_lines = ();

    open( my $fh, '<', $template_filename ) or do {
        my $logger = Log::Log4perl->get_logger();
        $logger->error("[DNS] Error opening $template_filename: $!");
        return;
    };
    @template_lines = <$fh>;
    close($fh) or do {
        my $logger = Log::Log4perl->get_logger();
        $logger->error("[DNS] Can not close $template_filename: $!");
    };
    return @template_lines;
}

=head2 incrementSerial

Returns a correctly incremented SOA serial number considering
the system date

=over 4

=item Parameter

 old serial number

=item Rückgabewert

correctly incremented SOA serial number

=back

=cut

sub incrementSerial
{
    my ($old_serial, $tm) = @_;
    if ( !defined $old_serial ) {
        $old_serial = "0000000000";
    }

    $tm ||= localtime;
    my $new_serial_prefix =
      sprintf( "%03d%03d", $tm->year, $tm->yday );

    my $increm = 0;
    if ( $new_serial_prefix eq substr( $old_serial, 0, 6 ) ) {
        $increm = substr( $old_serial, 6, 4 ) + 1;

        if ($increm > 9999) {
            my $logger = Log::Log4perl->get_logger();
            $logger->warn("[DNS] max increments for day and zone reached. DNS will not be updated anymore today!");
            $increm = 9999;
        }
    } 
    return sprintf( "%s%04d", $new_serial_prefix, $increm );
}

=head2 replaceSerial

Replace the DNS serial number in a zone file given as a reference
to a array of zone file lines.

=over 4

=item Parameter

 - reference to an array consisting of zone file lines which must
   contain a SOA record. The serial number must be found in the
   SOA line itself or in the next line.
 - new serial number

=item Return

returns nothing, the serial is replaced in the lines that the
passed array reference points to.

=back

=cut

sub replaceSerial
{
    my ( $zone_array_ref, $new_serial ) = @_;

    my $soa_line_rex = qr/SOA/ox;

    # find the line containing the serial number
    # normally it is the same line containing SOA or the next line
    my $line_number = 0;
    foreach my $line ( @{$zone_array_ref} ) {
        last if $line =~ /$soa_line_rex/;
        $line_number++;
    }

    # check if the SOA line was really found
    if ( $zone_array_ref->[$line_number] =~ /$soa_line_rex/ ) {

        #print $zone_array_ref->[$line_number]; # debug
        #print $zone_array_ref->[$line_number+1]; # debug

        # for now I assume the serial number is in the next line behind
        # the SOA line
        if ( not defined $new_serial ) {
            my $tm = localtime;

            $new_serial = sprintf( "%03d%03d%04d",
                $tm->year,
                $tm->yday,
                0 );
        }
        $zone_array_ref->[ $line_number + 1 ] =~ s/\d{10}/$new_serial/gs;
    }
    return;
}

=head2 zoneChanged

Comparison of two DNS zones, where one is given as an array consisting
of zone file lines and a path to a zone file on disk

=over 4

=item Parameter

 - reference to an array consisting of zone file lines
 - path to zone file on disk

=item Return

1 (true) if the zone changed , 0 (false) if no change detected

=back

=cut

sub zoneChanged
{
    my %param = (
        'zone_lines_aref' => undef,
        'zonefile_path'   => undef,
        @_,
    );

    my $zone_lines_ref = $param{'zone_lines_aref'};
    my $old_zone_file  = $param{'zonefile_path'};

    my $logger = Log::Log4perl->get_logger();

    open( my $old_zone_fh, '<', $old_zone_file ) or do {
        $logger->warn("[DNS] error opening $old_zone_file for compare: $!");
        return 1;
    };
    my @old_zone_lines = <$old_zone_fh>;
    close($old_zone_fh) or do {
        $logger->warn("[DNS] Can not close $old_zone_file: $!");
    };
    my $zone_changed = !stringArrayCompare( \@old_zone_lines, $zone_lines_ref );
    return $zone_changed;
}

=head2 backupZoneFile

Creates a backup copy of a zone file with .bak as extension

=over 4

=item Parameter

 - absolute path to zone file on disk

=item Return

1 (true) if the file could be copied or does not exist,
0 (false) on error

=back

=cut

sub backupZoneFile
{
    my ($zonefile_name) = @_;
    my $logger          = Log::Log4perl->get_logger();
    my $backup_filename = $zonefile_name . '.bak';

    # check, if old zone file exists and is readable
    if ( -r $zonefile_name ) {
        if ( !copy( $zonefile_name, $backup_filename ) ) {
            $logger->warn( "[DNS] error creating backup copy of "
                  . $zonefile_name
                  . ": $!" );
            return 0;
        }
    }
    return 1;
}

=head2 writeZoneFile

Write zone file

=over 4

=item Parameter

 - absolute path to zone file on disk
 - reference to array containing the lines of the zone file

=item Return

1 (true) if the file could be written
0 (false) on error

=back

=cut

sub writeZoneFile
{
    my %param = (
        'zone_lines_aref' => undef,
        'zonefile_path'   => undef,
        'config'          => undef,
        @_,
    );
    my $zonefile_path = $param{'zonefile_path'};
    my $zone_lines    = $param{'zone_lines_aref'};
    my $config        = $param{'config'};

    my $logger = Log::Log4perl->get_logger();

    my $zonefile_dir         = dirname($zonefile_path);
    my $zonefile_temp_suffix = $config->getWithDefault( 'LDAP2DNS_ZONEFILE_TEMP_SUFFIX', 'tmp' );

    $logger->debug("[DNS] writeZoneFile: calling checkDir with $zonefile_dir");
    if ( !checkDir($zonefile_dir) ) {
        return 0;
    }

    if ( $config->isEnabled( 'LDAP2DNS_NAMED_CHECKZONE_ACTIVE' ) ) {
        $zonefile_path = $zonefile_path . "." . $zonefile_temp_suffix;
    }

    open( my $out_fh, '>', $zonefile_path ) or do {
        $logger->error("[DNS] error opening $zonefile_path for writing: $!");
        return 0;
    };

    foreach my $line ( @{$zone_lines} ) {
        print $out_fh $line;
    }
    close($out_fh) or do {
        $logger->error("[DNS] Can not close $zonefile_path: $!");
    };
    return 1;
}

=head2 checkZoneFile

Check zone file

=over 4

=item Parameter

 - absolute path to zone file on disk
 - zone name
 - config

=item Return

1 (true) if the zone file is valid
0 (false) on error

=back

=cut

sub checkZoneFile
{
    my %param = (
        'zonefile_path' => undef,
        'zone_name'     => undef,
        'config'        => undef,
        @_,
    );

    my $zonefile_path = $param{'zonefile_path'};
    my $zone_name     = $param{'zone_name'};
    my $config        = $param{'config'};

    my $named_checkzone_bin     = $config->get( 'LDAP2DNS_NAMED_CHECKZONE_BIN' );
    my $named_checkzone_options =
      $config->get( 'LDAP2DNS_NAMED_CHECKZONE_OPTIONS' );
    my $named_checkzone_outfile =
      $config->get( 'LDAP2DNS_NAMED_CHECKZONE_OUTFILE' );
    my $named_check_outfile_suffix =
      $config->get( 'LDAP2DNS_NAMED_CHECK_OUTFILE_SUFFIX' );
    my $zonefile_temp_suffix = $config->get( 'LDAP2DNS_ZONEFILE_TEMP_SUFFIX' );
    my $zonefile_temp_path   = $zonefile_path . "." . $zonefile_temp_suffix;

    $named_checkzone_outfile =
        $named_checkzone_outfile . "."
      . $zone_name . "."
      . $named_check_outfile_suffix;

    my $logger = Log::Log4perl->get_logger();

    my $named_checkzone_ret =
      system(
        "$named_checkzone_bin $named_checkzone_options -c IN $zone_name $zonefile_temp_path > $named_checkzone_outfile 2>&1"
      );

    # log syntax check error message
    if ( $named_checkzone_ret != 0 ) {
        $logger->fatal(
            "[DNS] named zone syntax check for $zonefile_path (zone: $zone_name) failed, keeping old zonefile. See $named_checkzone_outfile for details"
        );
        return 0;
    }

    if ( !move( $zonefile_temp_path, $zonefile_path ) ) {
        $logger->error( "[DNS] could not move $zonefile_temp_path to "
              . $zonefile_path
              . " : $!" );
        return 0;
    }
    return 1;
}

=head2 checkZoneConfigFile

Check zone config file

=over 4

=item Parameter

 - config

=item Return

1 (true) if the zone config file is valid
0 (false) on error

=back

=cut

sub checkZoneConfigFile
{
    my %param = (
        'config' => undef,
        @_,
    );

    my $config = $param{'config'};

    my $zone_config_file_name;
    if ( isPrimary($config) ) {
        $zone_config_file_name = $config->get( 'LDAP2DNS_ZONECONFNAME' );
    } else {
        $zone_config_file_name = $config->get( 'LDAP2DNS_SLAVECONFNAME' );
    }
    my $named_checkconf_bin     = $config->get( 'LDAP2DNS_NAMED_CHECKCONF_BIN' );
    my $named_checkconf_options =
      $config->get( 'LDAP2DNS_NAMED_CHECKCONF_OPTIONS' );
    my $named_checkconf_outfile =
      $config->get( 'LDAP2DNS_NAMED_CHECKCONF_OUTFILE' );
    my $named_check_outfile_suffix =
      $config->get( 'LDAP2DNS_NAMED_CHECK_OUTFILE_SUFFIX' );
    my $zonefile_temp_suffix = $config->get( 'LDAP2DNS_ZONEFILE_TEMP_SUFFIX' );
    my $zone_config_file_temp_name =
      $zone_config_file_name . "." . $zonefile_temp_suffix;

    $named_checkconf_outfile =
        $named_checkconf_outfile . "."
      . basename($zone_config_file_name) . "."
      . $named_check_outfile_suffix;

    my $logger = Log::Log4perl->get_logger();

    my $named_checkconf_ret =
      system(
        "$named_checkconf_bin $named_checkconf_options $zone_config_file_temp_name > $named_checkconf_outfile 2>&1"
      );

    # log syntax check error message
    if ( $named_checkconf_ret != 0 ) {
        $logger->fatal(
            "[DNS] named config syntax check for $zone_config_file_name failed, keeping old config. See $named_checkconf_outfile for details"
        );
        return 0;
    }

    if ( !move( $zone_config_file_temp_name, $zone_config_file_name ) ) {
        $logger->error( "[DNS] could not move $zone_config_file_temp_name to "
              . $zone_config_file_name
              . " : $!" );
        return 0;
    }
    return 1;
}

sub getSubdomainList
{
    my %param = (
        'location_dn' => undef,
        'ldap_cache'  => undef,
        @_,
    );
    my $location_dn   = $param{'location_dn'};
    my $ldap_cache    = $param{'ldap_cache'};
    my @location_list = $ldap_cache->getAllLocationDns($location_dn);
    my @subdomains    = ();
    foreach my $i (@location_list) {

        # base ends with current base
        if ( $i =~ /.$location_dn$/i ) {
            my $valid_dns_name = ldap_dn2dnsname($i);
            if ($valid_dns_name) {
                push( @subdomains, $valid_dns_name );
            }
        }
    }
    return \@subdomains;
}

sub generateSubdomainNSRecords
{
    my %param = (
        'subdomains_aref' => undef,
        'config'          => undef,
        @_,
    );
    my $subdomains_aref = $param{'subdomains_aref'};
    my $config          = $param{'config'};
    my $dns_master_host = $config->get( 'LDAP2DNS_DNSMASTERHOST' );
    my @zone_line       = ( "\n", ";\n", "; subdomain nameservers\n", ";\n" );

    foreach my $i ( @{$subdomains_aref} ) {
        push( @zone_line, "$i.\tIN  NS\t$dns_master_host.\n" );
    }
    return @zone_line;
}

=head2 analyzeDHCPRange

Analyze a given DHCP range string and return prefix
or range without prefix depending on call context.

=over 4

=item Parameter

DHCP range string like used in the dbkDhcpRange attribute in the
format "IP, IP" (example: "10.201.5.1, 10.201.5.254")

=item Rückgabewert

B<scalar context:>
Returns prefix (first 3 octets) of DHCP range

B<array context:>
Returns an array consisting of the first and last number of
DHCP range without prefix. For example if the function was
called to return an array (array context) and the given range
is "10.201.4.1, 10.201.4.254" the array (1, 254) will be
returned.

=back

=cut

sub analyzeDHCPRange
{
    my ($ip_range) = @_;

    my $range_prefix;
    my @range;

    my $logger = Log::Log4perl->get_logger();

    if ( defined $ip_range ) {

        # remove all whitespace
        $ip_range =~ s/\s//g;
        my @ip_range = split( /,/, $ip_range );

        # check if first 3 octets are identical
        if ( $#ip_range < 1 ) {
            $logger->warn( "[DHCP] ip_range has not enough values in comma "
                  . "separated list: $ip_range" );
        }
        my @ip1       = split( /\./, $ip_range[0] );
        my @ip2       = split( /\./, $ip_range[1] );
        my $ip_prefix = join( ".",   @ip1[ 0 .. 2 ] );
        if ( $ip_prefix eq join( ".", @ip2[ 0 .. 2 ] ) ) {
            $range_prefix = $ip_prefix;
            @range = ( $ip1[3], $ip2[3] );
        }

    } else {
        $logger->warn("[DHCP] analyzeDHCPRange() called with undef");

        #show_callstack();
    }

    if (wantarray) {
        return @range;
    } else {
        return $range_prefix;
    }
}

=head2 getRangeRevZone

get the reverse zone name for a given IP range.
Assumption: the range is within a class C net (/24)

=over 4

=item Parameter

DHCP range string like used in the dbkDhcpRange attribute in the
format "IP, IP" (example: "10.201.5.1, 10.201.5.254")

=item Return

reverse zone name

=back

=cut

sub getRangeRevZone
{
    my ($dhcp_range) = @_;

    my $logger = Log::Log4perl->get_logger();

    $logger->debug("[DHCP] called with $dhcp_range");

    # convert IP range to be Net::IP compatible
    my $ip_range = $dhcp_range;
    $ip_range =~ s/,/-/g;
    $ip_range =~ s/\s+$//;
    $ip_range =~ s/^\s+//;

    my $range            = SmartClient::IP->new($ip_range);
    if (!$range) {
        $logger->warn("[LDAP-DATA] \"$ip_range\" is not valid to create a SmartClient::IP object.");
        return;
    }
    my $first_ip         = $range->first();
    my $first_ip_obj     = SmartClient::IP->new( $first_ip, '255.255.255.255' );
    my ($range_rev_zone) = $first_ip_obj->getReverseZones();
    return $range_rev_zone;
}

sub generateDHCPRangeRecord
{
    my %param = (
        'location_dn'       => undef,
        'ldap_cache'        => undef,
        'reverse_zone_flag' => 0,
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $ldap_cache  = $param{'ldap_cache'};
    my $rev_flag    = $param{'reverse_zone_flag'};
    my $logger      = Log::Log4perl->get_logger();

    my @zone_line = ();
    my $ip_prefix;
    my @num_range;

    $logger->debug("[DHCP] generateDHCPRangeRecord called with $location_dn");
    $logger->debug("[DHCP] generateDHCPRangeRecord rev_flag: $rev_flag");
    if ( !$ldap_cache->hasEntry($location_dn) ) {
        $logger->warn(
            "[DHCP,LDAP-DATA] no location entry found for $location_dn");
        return;
    }
    my $location_entry = $ldap_cache->getEntry($location_dn);
    my $ip_range       = $location_entry->get_value('dbkDhcpRange');

    if ( defined $ip_range ) {
        $ip_prefix = analyzeDHCPRange($ip_range);
        @num_range = analyzeDHCPRange($ip_range);

        if ( !defined $ip_prefix ) {
            $logger->warn( "[DHCP] first 3 octets in dbkDhcpRange "
                  . $location_dn
                  . " of $ip_range different" );
        } elsif ($rev_flag) {
            my $zone_name = $ldap_cache->getZoneNameForLocation($location_dn);
            push( @zone_line,
                    '$GENERATE '
                  . $num_range[0] . "-"
                  . $num_range[1] . "\t" . '$'
                  . "\tPTR\t" . 'dhcp$' . "."
                  . $zone_name
                  . ".\n" );
        } else {
            push( @zone_line,
                    '$GENERATE '
                  . $num_range[0] . "-"
                  . $num_range[1] . "\t" . 'dhcp$'
                  . "\t\tA\t"
                  . $ip_prefix . '.$'
                  . "\n" );

        }
    }
    return @zone_line;
}

=head2 generateIpDeviceRecords

Generate DNS records for IP devices (Workstations, Servers, Printers etc.)

=over 4

=item Parameter


=item Return


=back

=cut

sub generateIpDeviceRecords
{
    my %param = (
        'location_dn' => undef,
        'ldap_cache'  => undef,
        'zone_name'   => undef,
        @_,
    );

    my @zone_lines              = ();
    my $sc_cache                = $param{'ldap_cache'};
    my $location_dn             = $param{'location_dn'};
    my $zone_name               = $param{'zone_name'};
    my $location_data           = $sc_cache->getInternalEntryHash($location_dn);
    my $ip_device_objectclasses = $sc_cache->{'ip_device_objectclasses'};

    #print Data::Dumper->Dump([$location_data]); #debug
    my $logger = Log::Log4perl->get_logger();

    #my $ip_hash                 = $sc_cache->{'ip_addresses'};
    my %dns_names_hash;
    foreach my $objectclass ( @{$ip_device_objectclasses} ) {
        push( @zone_lines, "\n", ";\n", "; $objectclass records\n", ";\n" );
        my @ip_devices = $sc_cache->getLocationEntries(
            'base'             => $location_dn,
            'objectclass_name' => $objectclass
        );

        #print Data::Dumper->Dump([$ip_devices]); #debug
        foreach my $dn (@ip_devices) {
            if ( !$sc_cache->hasEntry($dn) ) {
                $logger->warn("[LDAP-DATA] no entry found for $dn");
                next;
            }
            my $entry    = $sc_cache->getEntry($dn);
            my @ip_addrs = $sc_cache->getIpAddress(
                entry_dn           => $dn,
                'all_ip_addresses' => 1
            );
            @ip_addrs = removeEmptyEntries(@ip_addrs);
            if ( !@ip_addrs ) {
                if ( lc( scalar $entry->get_value('dbkNoDNSName') ) eq 'true' )
                {
                    $logger->info(
                        "[LDAP-DATA] skipping $dn since dbkNoDNSName is true");
                    next;
                }
                $logger->error("[LDAP-DATA] No ip address for $dn found");
                next;
            }

            #print Data::Dumper->Dump([$sc_cache->{'entrycache'}->{$dn}]); #debug

            # write A record for first DNS name and CNAME for the rest
            my @dns_names = $entry->get_value('dbkDnsName');

            # apply dns name conformity
            @dns_names =
              grep { defined $_ } map { ldap_cn2dnsname($_) } @dns_names;
            my $first_dns_name =
              ldap_cn2dnsname( scalar $entry->get_value('cn') );
            if ( !$first_dns_name ) {
                if (@dns_names) {
                    $first_dns_name = shift @dns_names;
                } else {

                    # no valid cn and no valid dbkDnsName. skipping this dn
                    $logger->error(
                        "[DNS,LDAP-DATA] no valid dns name could be build for $dn"
                    );
                    next;
                }
            }

            # if cn exists as dns name too, remove it
            @dns_names = grep { $_ ne $first_dns_name } @dns_names;

            # prevent record with same name
            if ( $dns_names_hash{$first_dns_name} ) {
                $logger->warn(
                    "[DNS,LDAP-DATA] $first_dns_name in $zone_name already exists "
                      . "pointing to "
                      . $dns_names_hash{$first_dns_name} );
            } else {
                foreach my $ip_addr (@ip_addrs) {
                    push( @zone_lines,
                        sprintf( "%s\tIN\tA\t%s\n", $first_dns_name, $ip_addr )
                    );
                    if ( !exists $dns_names_hash{$first_dns_name} ) {
                        $dns_names_hash{$first_dns_name} = $ip_addr;
                    }
                }
            }
            foreach my $dns_name (@dns_names) {
                if ( $dns_names_hash{$dns_name} ) {
                    $logger->warn(
                        "[DNS,LDAP-DATA] $dns_name in $zone_name already exists "
                          . "pointing to "
                          . $dns_names_hash{$dns_name} );
                } else {
                    push(
                        @zone_lines,
                        sprintf(
                            "%s\tCNAME\t%s.%s.\n",
                            $dns_name, $first_dns_name, $zone_name
                        )
                    );
                    $dns_names_hash{$dns_name} =
                      $first_dns_name . '.' . $zone_name;
                }
            }
        }
    }
    return @zone_lines;
}

sub generateReverseIpRecords
{
    my %param = (
        'rev_zone_name' => undef,
        'ldap_cache'    => undef,
        @_,
    );

    my @zone_lines    = ();
    my $sc_cache      = $param{'ldap_cache'};
    my $rev_zone_name = $param{'rev_zone_name'};

    #my $last_octet_rex = s/.*\.//ox;

    my @rev_ip_list = $sc_cache->getIpAddressesForRevZone($rev_zone_name);
    foreach my $rev_ip (@rev_ip_list) {
        my @dns_names = $sc_cache->getDnsNamesForIp($rev_ip);
        ( my $last_octet = $rev_ip ) =~ s/.*\.//o;
        foreach my $dns_name (@dns_names) {
            push( @zone_lines,
                sprintf( "%s\tPTR\t%s.\n", $last_octet, $dns_name ) );
        }
    }

    return @zone_lines;
}

=head2 generateZone

Generates one DNS zone for a given location DN

=over 4

=item Parameter

 - location DN

=item Return

 - zone name if zone file was changed, empty string if
   zone file was not changed or undef if any error occured

=back

=cut

sub generateZone
{
    my %param = (
        'location_dn' => undef,
        'ldap_cache'  => undef,
        'config'      => undef,
        @_,
    );
    my $location_dn    = $param{'location_dn'};
    my $ldap_cache     = $param{'ldap_cache'};
    my $config         = $param{'config'};
    my @new_zone_lines = ();
    my $subdomains_aref;

    my $logger = Log::Log4perl->get_logger();

    # get the location entry from ldap cache
    $logger->debug("[DNS,LDAP-DATA] generateZone: location DN: $location_dn");
    if ( !$ldap_cache->hasEntry($location_dn) ) {
        $logger->warn(
            "[DNS,LDAP-DATA] no location entry for $location_dn found");
        return;
    }
    my $location_entry = $ldap_cache->getEntry($location_dn);

    # get zone name for location
    my $zone_name = $ldap_cache->getZoneNameForLocation($location_dn);
    if ( !$zone_name ) {
        $logger->warn(
            "[DNS,LDAP-DATA] no zone name for $location_dn found, will try to reconstruct it."
        );
        if ( $ldap_cache->setZoneNameForLocation($location_dn) ) {
            $zone_name = $ldap_cache->getZoneNameForLocation($location_dn);
        } else {
            $logger->warn(
                "[DNS,LDAP-DATA] could not reconstruct zone name for $location_dn"
            );
            return;
        }
    }

    # already stored in object by ScCache::getSerialForLocation
    #my $zone_name = $ldap_cache->{$location_dn}->{'zone_name'};
    $logger->info("[DNS] generating zone $zone_name");
    @new_zone_lines = readZoneTemplate( $config->get( 'LDAP2DNS_ZONETEMPLATE' ) );

    # write informational comment into array
    push( @new_zone_lines,
        "; Addresses\n",
        ";\n", "; Zone $zone_name generated by sc_ldap2dnsdhcpd.pl\n" );

    # generate NS records for subdomains
    $subdomains_aref = getSubdomainList(
        location_dn => $location_dn,
        ldap_cache  => $ldap_cache
    );

    # generate NS records for subdomains
    push(
        @new_zone_lines,
        generateSubdomainNSRecords(
            subdomains_aref => $subdomains_aref,
            config          => $config
        )
    );

    # build GENERATE line for dbkDhcpRange
    push(
        @new_zone_lines,
        generateDHCPRangeRecord(
            location_dn => $location_dn,
            ldap_cache  => $ldap_cache
        )
    );

    # build records for workstations, printers, servers and services
    push(
        @new_zone_lines,
        generateIpDeviceRecords(
            location_dn => $location_dn,
            ldap_cache  => $ldap_cache,
            zone_name   => $zone_name
        )
    );

    # insert predefined Zone extensions from ldap
    push(
        @new_zone_lines,
        generateZoneExtensions(
            location_dn => $location_dn,
            ldap_cache  => $ldap_cache
        )
    );

    my $zone_changed = writeZone(
        location_dn     => $location_dn,
        ldap_cache      => $ldap_cache,
        zone_name       => $zone_name,
        zone_lines_aref => \@new_zone_lines,
        config          => $config
    );
    if ($zone_changed) {
        return $zone_name;
    }

    # zone did not change
    return;
}

sub generateRevZones
{
    my %param = (
        'location_dn' => undef,
        'ldap_cache'  => undef,
        'config'      => undef,
        @_,
    );
    my $location_dn    = $param{'location_dn'};
    my $ldap_cache     = $param{'ldap_cache'};
    my $config         = $param{'config'};
    my @rev_zones      = ();
    my @new_zone_lines = ();
    my @changed_zones  = ();
    my $range_rev_zone;
    my $location_entry;
    my $ip_range;

    my $logger = Log::Log4perl->get_logger();

    # get the location entry from ldap cache
    $logger->debug(
        "[DNS,LDAP-DATA] generateRevZones: location DN: $location_dn");
    if ( !$ldap_cache->hasEntry($location_dn) ) {
        $logger->warn(
            "[DNS,LDAP-DATA] No entry found for $location_dn in generateRevZones"
        );
        return;
    }
    $location_entry = $ldap_cache->getEntry($location_dn);
    if ( $location_entry->exists('dbkDhcpRange') ) {
        $ip_range       = $location_entry->get_value('dbkDhcpRange');
        $range_rev_zone = getRangeRevZone($ip_range);
    }

    @rev_zones =
      $ldap_cache->getRevZonesForLocation( location_dn => $location_dn );
    $logger->debug(
        "[DNS,LDAP-DATA] found rev zones @rev_zones for $location_dn");

    foreach my $rev_zone_name (@rev_zones) {
        @new_zone_lines = ();
        @new_zone_lines =
          readZoneTemplate( $config->get( 'LDAP2DNS_ZONETEMPLATE' ) );

        # write informational comment into array
        push( @new_zone_lines,
            "; Addresses\n",
            ";\n", "; Zone $rev_zone_name generated by sc_ldap2dnsdhcpd.pl\n" );

        # check if the dhcp range is within this reverse zone
        # and build the $GENERATE line if applicable
        if ( $range_rev_zone && ( $rev_zone_name eq $range_rev_zone ) ) {
            if (
                !$ldap_cache->isLocation(
                    'dn'           => $location_dn,
                    'with_network' => 1
                )
              )
            {
                $logger->info(
                    "[DNS] will skip $location_dn, since it is not an network location"
                );

                #        return;
            } else {
                push(
                    @new_zone_lines,
                    generateDHCPRangeRecord(
                        location_dn       => $location_dn,
                        ldap_cache        => $ldap_cache,
                        reverse_zone_flag => 1
                    )
                );
            }
        }
        push(
            @new_zone_lines,
            generateReverseIpRecords(
                rev_zone_name => $rev_zone_name,
                ldap_cache    => $ldap_cache
            )
        );
        my $zone_changed = writeZone(
            location_dn     => $location_dn,
            ldap_cache      => $ldap_cache,
            zone_name       => $rev_zone_name,
            zone_lines_aref => \@new_zone_lines,
            config          => $config
        );
        if ($zone_changed) {
            push( @changed_zones, $rev_zone_name );
        }
    }
    return @changed_zones;
}

sub generateZoneConfigFile
{
    my %param = (
        'ldap_cache' => undef,
        'config'     => undef,
        @_,
    );

    my $sc_cache = $param{'ldap_cache'};
    my $config   = $param{'config'};

    my $rev_zones_hash = $sc_cache->{'rev_zones'};

    my $logger = Log::Log4perl->get_logger();

    my $zone_config_file_name;
    if ( isPrimary($config) ) {
        $zone_config_file_name = $config->get( 'LDAP2DNS_ZONECONFNAME' );
    } else {
        $zone_config_file_name = $config->get( 'LDAP2DNS_SLAVECONFNAME' );
    }

    my $slave_zone_path = $config->get( 'LDAP2DNS_SLAVEZONEPATH' );

    my $zonefile_temp_suffix = $config->getWithDefault( 'LDAP2DNS_ZONEFILE_TEMP_SUFFIX', 'tmp' );

    my ($zone_type, $dns_master_ip, $dns_slave_ip) = getDnsMasterSlaveSetup($config);

    if ( $config->isEnabled( 'LDAP2DNS_NAMED_CHECKCONF_ACTIVE' ) ) {
        $zone_config_file_name =
          $zone_config_file_name . "." . $zonefile_temp_suffix;
    }

    $logger->info("[DNS] writing zone config file $zone_config_file_name");
    open( my $zone_fh, '>', $zone_config_file_name ) or do {
        $logger->error("[DNS] opening $zone_config_file_name failed: $!");
        return;
    };
    print $zone_fh "# include config file generated by sc_ldap2dnsdhcpd.pl\n\n";

    foreach my $location_dn ( $sc_cache->getAllLocationDns() ) {
          generateZoneConfigFileForLocation(
            location        => $location_dn,
            sc_cache        => $sc_cache,
            fh              => $zone_fh,
            slave_zone_path => $slave_zone_path,
            master_ip       => $dns_master_ip,
            slave_ip        => $dns_slave_ip,
            zone_type       => $zone_type,
          );
    }

    foreach my $rev_zone_name ( keys %{$rev_zones_hash} ) {
        my $zone_file_path;
        if ( $zone_type eq 'master' ) {
            $zone_file_path = $sc_cache->getRevZoneFileName($rev_zone_name);
        }
        print $zone_fh "zone \"$rev_zone_name\" IN {\n";
        print $zone_fh "\ttype $zone_type;\n";
        if ( $zone_type eq 'slave' ) {
            $zone_file_path = $slave_zone_path . '/' . $rev_zone_name;
            print $zone_fh "\tmasters { " . $dns_master_ip . "; };\n";
        } else {
            print $zone_fh "\tallow-update { none; };\n";
            if ($dns_slave_ip) {
                print $zone_fh "\tallow-transfer { " . $dns_slave_ip . "; };\n";
                print $zone_fh "\tnotify yes;\n";
                print $zone_fh "\talso-notify { " . $dns_slave_ip . "; };\n";
            }
        }
        print $zone_fh "\tfile \"$zone_file_path\";\n";
        print $zone_fh "};\n\n";
    }
    close($zone_fh) or do {
        $logger->error("[DNS] Can not close zone config file $zone_config_file_name: $!");
    };
    return 1;
}

sub getDnsMasterSlaveSetup
{
    my ($config) = @_;
    if ( $config->get('LDAP2DNSDHCPD_HA_MODE') eq 'ha_mode_secondary' ) {
        return (
            'slave',
            $config->get('LDAP2DNS_DNSMASTERIP'),
            $config->get('LDAP2DNS_DNSSLAVEIP')
        );
    }
    return ( 'master', undef, $config->get( 'LDAP2DNS_DNSSLAVEIP', 1 ) );
}

sub filterValidIps
{
    my ( $reason, @ips ) = @_;
    my $logger = Log::Log4perl->get_logger();
    my $result = EMPTY_STR;
    foreach my $ip (@ips) {
        if ( SmartClient::IP->new($ip) ) {
            $result .= $ip . "; ";
        } else {
            $logger->warn(
                "[DNS] IP " . $ip . " for $reason option isn't valid." );
        }
    }
    return $result;
}

sub generateZoneConfigFileForLocation
{
    my %params = (
        location        => undef,
        sc_cache        => undef,
        fh              => undef,
        slave_zone_path => undef,
        master_ip       => undef,
        slave_ip        => undef,
        zone_type       => undef,
        allow_transfer  => undef,
        also_notify     => undef,
        @_
    );
    my $location_dn       = $params{location};
    my $sc_cache          = $params{sc_cache};
    my $zone_fh           = $params{fh};
    my $slave_zone_path   = $params{slave_zone_path};
    my $dns_master_ip     = $params{master_ip};
    my $dns_slave_ip      = $params{slave_ip};
    my $zone_type         = $params{zone_type};
    my $logger            = Log::Log4perl->get_logger();

    my $location_entry    = $sc_cache->getEntry($location_dn);
    my $allowZoneTransfer =
      filterValidIps( 'allow-transfer',
        $location_entry->get_value("dbkDnsAllowZoneTransfer") );
    my $alsoNotify =
      filterValidIps( 'also-notify',
        $location_entry->get_value("dbkDnsAlsoNotify") );

    my @zone_names         = ();
    my $location_zone_name = $sc_cache->getLocationZoneName($location_dn);
    if ( !$location_zone_name ) {
        $logger->warn(
            "[DNS,LDAP-DATA] Could not get zone name for $location_dn. So no zones for this location will be written."
        );
        next;
    }
    push( @zone_names, $location_zone_name );
    my $zone_file_path;
    if ( $zone_type eq 'master' ) {
        $zone_file_path = $sc_cache->getLocationZoneFileName($location_dn);
    }
    push( @zone_names, $sc_cache->getLocationAliasZones($location_dn) );
    foreach my $zone_name (@zone_names) {
        print $zone_fh "zone \"$zone_name\" IN {\n";
        print $zone_fh "\ttype " . $zone_type . ";\n";

        if ( $zone_type eq 'slave' ) {
            $zone_file_path = $slave_zone_path . '/' . $zone_name;
            print $zone_fh "\tmasters { " . $dns_master_ip . "; };\n";
        } else {
            if ($dns_slave_ip) {

                print $zone_fh "\tallow-transfer { "
                  . $dns_slave_ip . "; "
                  . $allowZoneTransfer . "};\n";
                print $zone_fh "\tnotify yes;\n";
                print $zone_fh "\talso-notify { "
                  . $dns_slave_ip . "; "
                  . $alsoNotify . "};\n";
            }
            print $zone_fh "\tallow-update { none; };\n";
        }
        print $zone_fh "\tfile \"$zone_file_path\";\n";
        print $zone_fh "};\n\n";
    }

    return;
}

sub writeZone
{
    my %param = (
        'location_dn'     => undef,
        'ldap_cache'      => undef,
        'zone_name'       => undef,
        'zone_lines_aref' => undef,
        'config'          => undef,
        @_,
    );
    my $location_dn     = $param{'location_dn'};
    my $ldap_cache      = $param{'ldap_cache'};
    my $zone_name       = $param{'zone_name'};
    my $zone_lines_aref = $param{'zone_lines_aref'};
    my $config          = $param{'config'};

    my $rev_zone_rex = qr/arpa$/ox;
    my $old_serial;
    my $zonefile_path;

    my $logger = Log::Log4perl->get_logger();

    # get the location entry from ldap cache
    $logger->debug("[DNS] writeZone: zone name: $zone_name");

    # get old serial from cache
    #my $old_serial = $ldap_cache->{$location_dn}->{'serial'};
    if ( $zone_name =~ /$rev_zone_rex/ ) {
        $old_serial =
          $ldap_cache->getSerialForRevZone( rev_zone_name => $zone_name );
        $zonefile_path = $ldap_cache->getRevZoneFileName($zone_name);
    } else {
        $old_serial =
          $ldap_cache->getSerialForLocation( location_dn => $location_dn );
        $zonefile_path = $ldap_cache->getLocationZoneFileName($location_dn);
    }

    # check if the zone changed, this is done by replacing the
    # serial in new zone held in the array @new_zone_lines and
    # then comparing it with the zone file on disk
    replaceSerial( $zone_lines_aref, $old_serial );
    $logger->debug("[DNS] calling zoneChanged with $zonefile_path");
    if (
        zoneChanged(
            zone_lines_aref => $zone_lines_aref,
            zonefile_path   => $zonefile_path
        )
      )
    {

        # increment serial number
        my $new_serial = incrementSerial($old_serial);
        if ( length($new_serial) > 10 ) {
            $logger->error(
                "[DNS] serial number too big in $zonefile_path (this means more than 9999 changes in this zone today)"
            );
        }
        $logger->debug( '[DNS] serial old: '
              . ( $old_serial || 'NO OLD SERIAL' )
              . ' new: '
              . ( $new_serial || 'NO NEW SERIAL' ) . ' for '
              . ( $zonefile_path || 'NO ZONE FILE' ) );

        # replace serial number in generated array with incremented
        # serial number
        replaceSerial( $zone_lines_aref, $new_serial );

        # create backup copy of zone file
        backupZoneFile( $zonefile_path, $zone_name );

        # write the zone file
        $logger->debug("[DNS] calling writeZoneFile with $zonefile_path");
        writeZoneFile(
            zonefile_path   => $zonefile_path,
            zone_lines_aref => $zone_lines_aref,
            config          => $config
        );

        # check the zone config file
        if ( $config->isEnabled( 'LDAP2DNS_NAMED_CHECKZONE_ACTIVE' ) ) {
            $logger->info(
                "[DNS] calling checkZoneFile with $zonefile_path and zone $zone_name"
            );
            checkZoneFile(
                zonefile_path => $zonefile_path,
                zone_name     => $zone_name,
                config        => $config
            );
        }

        # store incremented serial in cache
        if ( $zone_name =~ /$rev_zone_rex/ ) {
            $ldap_cache->setSerialForRevZone(
                rev_zone_name => $zone_name,
                serial        => $new_serial
            );
        } else {
            $ldap_cache->setSerialForLocation(
                location_dn => $location_dn,
                serial      => $new_serial
            );
        }

        # zone was changed
        return 1;
    }

    # zone did not change
    return;
}

=head2 generateZoneExtensions

Generates DNS zone files for a given list of locations

=over 4

=item Parameter

 - location_dn: location DN
 - ldap_cache: SmartClient::ScCache object

=item Return

 - list of RRs

=back

=cut

sub generateZoneExtensions
{
    my %param = (
        'location_dn' => undef,
        'ldap_cache'  => undef,
        @_,
    );

    my $logger = Log::Log4perl->get_logger();

    my $attribute   = 'dbkDnsZoneExtension';
    my $sc_cache    = $param{'ldap_cache'};
    my $location_dn = $param{'location_dn'};

    if ( !$sc_cache->hasEntry($location_dn) ) {
        $logger->warn("[DNS,LDAP-DATA] no location found for $location_dn");
        return;
    }
    my $entry                = $sc_cache->getEntry($location_dn);
    my $dbkDnsZoneExtensions = $entry->get_value( $attribute, asref => 1 );
    my @zone_lines           = ( ";\n", "; zone extensions\n", ";\n" );

    foreach my $dbkDnsZoneExtension ( @{$dbkDnsZoneExtensions} ) {
        push( @zone_lines, $dbkDnsZoneExtension, "\n" );
        $logger->info(
            "[DNS] Attribut: " . $attribute . "  :  " . $dbkDnsZoneExtension );
    }

    push( @zone_lines, "\n" );

    return @zone_lines;
}

=head2 generateLocationZones

Generates DNS zone files for a given list of locations

=over 4

=item Parameter

 - changed_locations: list of location DNs
 - ldap_cache: SmartClient::ScCache object

=item Return

 - list of zone names with changes to the zone file

=back

=cut

sub generateLocationZones
{
    my %param = (
        'changed_locations' => undef,
        'ldap_cache'        => undef,
        'config'            => undef,
        @_,
    );

    my $changed_locations = $param{changed_locations};
    my $ldap_cache        = $param{ldap_cache};
    my $config            = $param{config};
    my @changed_zones     = ();
    my $logger            = Log::Log4perl->get_logger();
    $logger->info("[DNS] generateLocationZones was called");
    foreach my $location_dn ( @{$changed_locations} ) {
        $logger->debug(
            "[DNS] generateLocationZones calling generateZone with $location_dn"
        );
        my $changed_zone_name = generateZone(
            'location_dn' => $location_dn,
            'ldap_cache'  => $ldap_cache,
            'config'      => $config
        );
        if ($changed_zone_name) {
            push( @changed_zones, $changed_zone_name );
            push( @changed_zones,
                $ldap_cache->getLocationAliasZones($location_dn) );
        }
        my @changed_rev_zone_names = generateRevZones(
            'location_dn' => $location_dn,
            'ldap_cache'  => $ldap_cache,
            'config'      => $config
        );
        push( @changed_zones, @changed_rev_zone_names );
    }
    return @changed_zones;
}

sub reloadChangedDnsZones
{
    my %param = (
        'changed_zones_aref' => undef,
        'config'             => undef,
        @_,
    );

    my $changed_zones_aref = $param{'changed_zones_aref'};
    my $config             = $param{'config'};

    my $logger = Log::Log4perl->get_logger();

    my $number_zones_changed = @{$changed_zones_aref};
    if ( $number_zones_changed < 1 ) {
        return;
    }
    my @ndc_cmds;
    if ( $config->get( 'LDAP2DNS_BINDVERSION', 1 ) == 8 ) {
        @ndc_cmds = generateDnsReloadCmd( 'ndc', @{$changed_zones_aref} );
    } elsif ( $config->get( 'LDAP2DNS_BINDVERSION', 1 ) == 9 ) {
        my $rndc_cmd = $config->getWithDefault( 'LDAP2DNS_RNDC', '/usr/sbin/rndc' );
        @ndc_cmds = generateDnsReloadCmd( $rndc_cmd, @{$changed_zones_aref} );
    } else {
        $logger->warn( '[DNS] unknown bind version specified in config file: '
              . $config->get( 'LDAP2DNS_BINDVERSION' )
              . ' doing reload' );
        @ndc_cmds = ('/etc/init.d/named reload');
    }
    foreach my $ndc_cmd (@ndc_cmds) {
        $logger->info("[DNS] running $ndc_cmd");
        my $rc = system("$ndc_cmd 1>/dev/null 2>/tmp/ndc_reload_stderr.tmp");
        $logger->info("[DNS] $ndc_cmd returned $rc");
    }
    return;
}

sub generateDnsReloadCmd
{
    my ($cmd, @zones) = @_;
    return ("$cmd reconfig", map { "$cmd reload $_" } @zones);
}

sub reloadSlaveDns
{
    my %param = (
        'config' => undef,
        @_,
    );

    my $config = $param{'config'};

    my $logger     = Log::Log4perl->get_logger();
    my $ndc_reload = EMPTY_STR;
    if ( $config->get( 'LDAP2DNS_BINDVERSION' ) == 8 ) {
        $ndc_reload = 'ndc reconfig,ndc reload';
    } elsif ( $config->get( 'LDAP2DNS_BINDVERSION' ) == 9 ) {
        my $rndc_reload = $config->getWithDefault( 'LDAP2DNS_RNDC', '/usr/sbin/rndc' );
        $ndc_reload = $rndc_reload . ' reconfig,' . $rndc_reload . ' reload';
    } else {
        $logger->warn( "[DNS] unknown bind version specified in config file: "
              . $config->get( 'LDAP2DNS_BINDVERSION' )
              . " doing reload" );
        $ndc_reload = "/etc/init.d/named reload";
    }
    my @ndc_cmds = split( /,/, $ndc_reload );
    foreach my $ndc_cmd (@ndc_cmds) {
        $logger->info("[DNS] running $ndc_cmd");
        my $rc = system("$ndc_cmd 1>/dev/null 2>/tmp/ndc_reload_stderr.tmp");
        $logger->info("[DNS] $ndc_cmd returned $rc");
    }
    return;
}

sub writeDhcpHeader
{
    my ( $tmp_fh, $headerfile ) = @_;
    my $logger = Log::Log4perl->get_logger();

    # read all lines from the header file
    open( my $header_fh, '<', $headerfile ) or do {
        $logger->error("[DHCP] error opening DHCP header file $headerfile: $!");
        return;
    };
    my @headerlines = <$header_fh>;
    close($header_fh) or do {
        $logger->error("[DHCP] Can not close header file $headerfile: $!");
    };

    foreach my $line (@headerlines) {
        print $tmp_fh $line;
    }
    print $tmp_fh "\n# end of header\n\n";
    return 1;
}

=head2 convertDnsToDhcpDomainSearchList

convert a domainname plus base domainname and additional (server) domains to a string
conforming to RFC3397 for dhcp option 119 which is domain search list.

B<This is only useful, if the dhcp-client requests this option.>

=head3 Example

  my $search_list = convertDnsToDhcpDomainSearchList(
    'domain'          => $domain,
    'base_domain'     => 'debeka.de',             
    'additional_domains' => [ 'server', 'master' ]
  ); 

=over 4

=item Parameter

hash consisting of:

=over 4

=item C<domain>

dns name for the domain

=item C<base_domain>

dns name for the base domain

=item C<additional_domains>

reference to an array with all additional domains. These have to be relative to C<base_domain> and must 
be children of it. 

=back

=item Return

string representing RFC3397 conform domain search list
 
=back

=cut

sub convertDnsToDhcpDomainSearchList
{
    my %params = (
        'domain'             => undef,
        'base_domain'        => undef,
        'additional_domains' => [],
        @_
    );
    my $domain      = $params{'domain'};
    my $base_domain = $params{'base_domain'};

    if ($domain) {
        my $base_domain_string     = EMPTY_STR;
        my $base_domain_position   = 0;
        my @base_domain_components = ();
        if ( $base_domain and $base_domain ne $domain ) {
            @base_domain_components = split( /\./, $base_domain );
            $base_domain_string =
              convertDnsToDhcpDomainSearchList( 'domain' => $base_domain );

            # remove everything after the complete domain name including the ending marker \000
            $base_domain_string =~ s/\\000.*$//;
            $domain             =~ s/$base_domain$//i;
        }
        my $search_list = EMPTY_STR;
        my @domain_components;
        my $first_round = 1;
        @domain_components = split( /\./, $domain );
        do {
            foreach my $domain_component (@domain_components) {
                $search_list = sprintf( "%s\\%03o%s",
                    $search_list, length($domain_component),
                    $domain_component );
                if ($first_round) {
                    $base_domain_position += length($domain_component) + 1;
                }
            }
            if ($first_round) {

                # paste the full base domain into the search list
                $search_list = $search_list . $base_domain_string . '\\000';
                $first_round = 0;
            } else {

                # place a reference to the base domain
                $search_list = sprintf( "%s\\%03o\\%03o",
                    $search_list, 0xC0, $base_domain_position );
            }

            shift(@domain_components);
        } while (@domain_components);
        foreach my $additional_domain ( @{ $params{'additional_domains'} } ) {
            $search_list = sprintf(
                "%s\\%03o%s\\%03o\\%03o",
                $search_list, length($additional_domain),
                $additional_domain, 0xC0, $base_domain_position
            );
        }
        return $search_list;
    }
    return;
}

sub getDhcpSearchList
{
    my %param = (
        'location_dn' => undef,
        'sc_cache'    => undef,
        'config'      => undef,
        @_,
    );
    my $location_dn = $param{'location_dn'};
    my $sc_cache    = $param{'sc_cache'};
    my $config      = $param{'config'};
    my $logger      = Log::Log4perl->get_logger();

    my $dns_base = ldap_dn2dnsname( $config->get( 'ALL_LDAPBASE' ) );
    if ( !$dns_base ) {
        $logger->error(
            "[DHCP,LDAP-DATA] no dns base could be build for ALL_LDAPBASE="
              . $config->get( 'ALL_LDAPBASE' ) );
    }
    my $dns_domain = $sc_cache->getLocationZoneName($location_dn);
    if ( !$dns_domain ) {

        #there seems to be no location with this name
        $logger->warn(
            "[DHCP,DNS] no zone name for location $location_dn found");
        return;
    }
    my @dns_components = split( /\./, $dns_domain );
    my @search_list = ( join( ".", @dns_components ) );

    # remove first components of dns domain
    # until reaching the base domain.
    # add the subdomains to the searchlist
    #while ( length($dns_domain) >= length($dns_base) ) {
    #    push @search_list, $dns_domain;
    #    shift @dns_components;
    #    $dns_domain = join( ".", @dns_components );
    #}

    my $search =
      join( ONE_SPACE, @search_list ) . ONE_SPACE
      . $config->get( 'LDAP2DHCP_APPEND_SEARCHLIST' );

    return $search;
}

sub convertDhcpOptionsToHash
{
    my ($dhcp_opts_hash, $dhcp_opts) = @_;

    foreach my $dhcp_option ( @{$dhcp_opts} ) {
        if ( $dhcp_option =~ /(.+?)\s(.+)/ ) {
            $dhcp_opts_hash->{$1} = $2;
        }
    }
    return;
}

sub getDhcpOptions
{

    # retrieve and merge dhcp options
    # settings in location base take precedence
    my %param = (
        'location_dn' => undef,
        'sc_cache'    => undef,
        @_,
    );

    my $location_dn = $param{'location_dn'};
    my $sc_cache    = $param{'sc_cache'};

    my %dhcp_opts_hash;
    my @dhcp_options = ();

    #my $base_below   = ldap_shiftbase($location_dn);
    #TODO: if this really is a location, then a parent of a location will be a region
    #      for a region there are no referenceAttributes, so why search for them?
    my @dhcp_opts = ();

    #my @dhcp_opts = $sc_cache->getReferenceAttribute(
    #    entry_dn                 => $base_below,
    #    reference_attribute_name => 'dhcpOption'
    #);
    #if (@dhcp_opts) {
    #    convertDhcpOptionsToHash( \%dhcp_opts_hash, \@dhcp_opts );
    #}
    my $logger = Log::Log4perl->get_logger();
    if ( !$sc_cache->hasEntry($location_dn) ) {
        $logger->warn("[DHCP,LDAP-DATA] no location found for $location_dn");
        return;
    }
    my $location_entry = $sc_cache->getEntry($location_dn);
    if ( $location_entry->exists('dhcpOption') ) {
        @dhcp_opts = $location_entry->get_value('dhcpOption');
        convertDhcpOptionsToHash( \%dhcp_opts_hash, \@dhcp_opts );
    }
    foreach my $dhcp_opt_key ( keys %dhcp_opts_hash ) {
        push( @dhcp_options,
            $dhcp_opt_key . ONE_SPACE . $dhcp_opts_hash{$dhcp_opt_key} );
    }
    return @dhcp_options;
}

sub splitDhcpRange
{
    my ($tmprange) = @_;

    # remove blanks from range string
    $tmprange =~ s/\s+//g;
    my @range = split( /,/, $tmprange );
    return @range;
}

sub writeDhcpLeaseTime
{
    my %param = (
        'tmp_fh'              => undef,
        'config'              => undef,
        'boot_opt'            => undef,
        'dyn_ip'              => 0,
        'localboot_leasetime' => undef,
        'dyn_leasetime'       => undef,
        'netboot_leasetime'   => undef,
        'indent_level'        => undef,
        @_,
    );

    my $tmp_fh       = $param{'tmp_fh'};
    my $config       = $param{'config'};
    my $dyn_ip       = $param{'dyn_ip'};
    my $boot_opt     = lc( $param{'boot_opt'} );
    my $space_str;

    my $dyn_leasetime = $param{'dyn_leasetime'}
      || $config->get( 'LDAP2DHCP_DYNLEASETIME' );
    my $netboot_leasetime = $param{'netboot_leasetime'}
      || $config->get( 'LDAP2DHCP_NETBOOTLEASETIME' );
    my $localboot_leasetime = $param{'localboot_leasetime'}
      || $config->get( 'LDAP2DHCP_LOCALBOOTLEASETIME' );
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $space_str    = (INDENT_SPACER) x $param{'indent_level'};
    } else {
        $space_str = DEFAULT_SPACER;
    }

    if ( $boot_opt eq 'localboot' ) {

        # write lease-time entry for localboot
        print $tmp_fh $space_str
          . "default-lease-time "
          . $localboot_leasetime . ";\n";
        print $tmp_fh $space_str
          . "max-lease-time "
          . $localboot_leasetime . ";\n";
    } else {

        # write lease-time entry for netboot
        # $boot_opt is lan, wan or a name in this case, which
        # always means netboot
        # if dyn_ip is true, a dynamic range is defined and
        # the associated lease-time is used.
        if ($dyn_ip) {
            print $tmp_fh $space_str
              . "default-lease-time "
              . $dyn_leasetime . ";\n";
            print $tmp_fh $space_str
              . "max-lease-time "
              . $dyn_leasetime . ";\n";
        } else {
            print $tmp_fh $space_str
              . "default-lease-time "
              . $netboot_leasetime . ";\n";
            print $tmp_fh $space_str
              . "max-lease-time "
              . $netboot_leasetime . ";\n";
        }
    }
    return;
}

sub writeDhcpNextServer
{
    my %param = (
        'tmp_fh'       => undef,
        'config'       => undef,
        'boot_opt'     => undef,
        'indent_level' => undef,
        @_,
    );

    my $tmp_fh   = $param{'tmp_fh'};
    my $config   = $param{'config'};
    my $boot_opt = lc( $param{'boot_opt'} );
    my $space_str;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $space_str    = (INDENT_SPACER) x $param{'indent_level'};
    } else {
        $space_str = DEFAULT_SPACER;
    }
    my $next_server;

    if ( $boot_opt eq 'wan' ) {
        $next_server = 'tftp';
    } elsif ( $boot_opt eq 'lan' ) {

        # code for finding appropriate TFTP server in LAN
        # would have to be plugged in here
        $next_server = 'tftp';
    } elsif ( $boot_opt eq 'localboot' ) {
        $next_server = 'localboot';
    } else {
        $next_server = $boot_opt;
    }
    print $tmp_fh $space_str . "next-server ", $next_server, ";\n";
    return;
}

sub generatePxeLinkTarget
{

    # generate PXE link target from MAC address
    # PXE looks for a file consisting of the MAC address but
    # separated by "-" instead of ":" and prepended by "01-"
    # example: 00:30:05:23:7e:AE -> 01-00-30-05-23-7e-ae
    # (all lowercase)

    my $mac_addr = lc(shift);
    $mac_addr =~ s/:/-/go;
    return '01-' . $mac_addr;
}

sub createPxeConfigLink
{
    my %param = (
        'pxe_config_file' => undef,
        'config'          => undef,
        'entry_dn'        => undef,
        'sc_cache'        => undef,
        @_,
    );

    my $config = $param{'config'};

    # return immediately,
    # if generation of PXE config links is disabled in config.
    return if ( $config->isDisabled( 'LDAP2DHCP_CREATE_PXE_CFG_LINKS' ) );

    my $pxe_config_file = $param{'pxe_config_file'};
    my $entry_dn        = $param{'entry_dn'};
    my $sc_cache        = $param{'sc_cache'};

    my $pxe_config_dir = $config->get( 'ALL_PXE_CFG_DIR' );

    # append "/" if missing
    if ( !$pxe_config_dir =~ /\/$/o ) {
        $pxe_config_dir .= '/';
    }

    my $logger = Log::Log4perl->get_logger();
    if ( !$sc_cache->hasEntry($entry_dn) ) {
        $logger->warn("[DHCP,LDAP-DATA] no entry found for $entry_dn");
        return;
    }

    my $entry = $sc_cache->getEntry($entry_dn);

    if ( !$entry->exists('macAddress') ) {
        $logger->warn(
            "[DHCP] $entry_dn does not have macAddress, not creating PXE link");
        return;
    }

    my $pxe_config_filepath = $pxe_config_dir . $pxe_config_file;
    if ( !-f $pxe_config_filepath ) {
        $logger->warn("[DHCP] file $pxe_config_filepath does not exist");
        return;
    }

    my $mac_address = $entry->get_value('macAddress');
    my $link_target = generatePxeLinkTarget($mac_address);
    $logger->info(
        "[DHCP] trying to create pxe link $pxe_config_dir$link_target to $pxe_config_file"
    );
    my $symlink_rc =
      checkSymLink( $pxe_config_file, $pxe_config_dir . $link_target );

    if ($symlink_rc) {
        $logger->info(
            "[DHCP] create PXE link for $entry_dn to $pxe_config_file returned "
              . ( $symlink_rc || "0" ) );
    } else {
        $logger->error(
                "[DHCP] creating PXE link for $entry_dn to $pxe_config_file "
              . " in "
              . $config->get( 'ALL_PXE_CFG_DIR' )
              . " failed: $!" );
    }

    return $symlink_rc;
}

sub removePxeConfigLink
{
    my %param = (
        'config'   => undef,
        'entry_dn' => undef,
        'sc_cache' => undef,
        @_,
    );

    my $config = $param{'config'};
    return if ( $config->isDisabled( 'LDAP2DHCP_CREATE_PXE_CFG_LINKS' ) );

    my $entry_dn = $param{'entry_dn'};
    my $sc_cache = $param{'sc_cache'};

    my $pxe_config_dir = $config->get( 'ALL_PXE_CFG_DIR' );

    # append "/" if missing
    if ( !$pxe_config_dir =~ /\/$/o ) {
        $pxe_config_dir .= '/';
    }

    my $logger = Log::Log4perl->get_logger();
    if ( !$sc_cache->hasEntry($entry_dn) ) {
        $logger->warn("[DHCP,LDAP-DATA] no entry found for $entry_dn");
        return;
    }

    my $entry = $sc_cache->getEntry($entry_dn);

    if ( !$entry->exists('macAddress') ) {
        $logger->warn(
            "[DHCP,LDAP-DATA] $entry_dn does not have macAddress, not removing PXE link"
        );
        return;
    }

    my $mac_address = $entry->get_value('macAddress');
    my $link_target = generatePxeLinkTarget($mac_address);
    my $symlink_rc  = checkRemoveSymLink( $pxe_config_dir . $link_target );
    return $symlink_rc;
}

sub writeDhcpTftpBootOptions
{
    my %param = (
        'tmp_fh'       => undef,
        'config'       => undef,
        'entry_dn'     => undef,
        'boot_opt'     => undef,
        'sc_cache'     => undef,
        'indent_level' => undef,
        @_,
    );

    my $tmp_fh   = $param{'tmp_fh'};
    my $config   = $param{'config'};
    my $entry_dn = $param{'entry_dn'};
    my $boot_opt = lc( $param{'boot_opt'} );
    my $sc_cache = $param{'sc_cache'};
    my $space_str;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $space_str    = (INDENT_SPACER) x $param{'indent_level'};
    } else {
        $space_str = DEFAULT_SPACER;
    }

    my $boot_opts_value;
    my $pxe_config_file;

    my $logger = Log::Log4perl->get_logger();

    my ($ref_boot_options_dn) = $sc_cache->getReferenceAttribute(
        entry_dn                 => $entry_dn,
        reference_attribute_name => 'scRefBootOptionsDn'
    );
    if ( !$ref_boot_options_dn ) {
        $logger->warn("[DHCP] could not get scRefBootOptionsDn for $entry_dn");
        return;
    }
    my $ref_boot_opts_entry = $sc_cache->getEntry( lc($ref_boot_options_dn) );
    if ( !$ref_boot_opts_entry ) {
        $logger->warn( "[DHCP] could not get reference boot options entry "
              . "$ref_boot_options_dn for $entry_dn from cache" );
        return;
    }

    my $entry     = $sc_cache->getEntry($entry_dn);
    my $is_device = $sc_cache->isDevice($entry);
    if ( $boot_opt eq 'localboot' ) {
        $boot_opts_value =
          $ref_boot_opts_entry->get_value('dbkDhcpOptionsLocal');
        if ($is_device) {
            $pxe_config_file =
              $ref_boot_opts_entry->get_value('scPxeCfgFileLocal');
        }
    } else {
        $boot_opts_value =
          $ref_boot_opts_entry->get_value('dbkDhcpOptionsRemote');
        if ($is_device) {
            $pxe_config_file =
              $ref_boot_opts_entry->get_value('scPxeCfgFileRemote');
        }
    }
    print $tmp_fh $space_str . $boot_opts_value, "\n";

    if ($is_device) {
        if ($pxe_config_file) {
            my $symlink_rc = createPxeConfigLink(
                pxe_config_file => $pxe_config_file,
                entry_dn        => $entry_dn,
                config          => $config,
                sc_cache        => $sc_cache
            );
        } else {

            #$logger->warn("could not get PXE config file from $ref_boot_options_dn "
            #    . " for $entry_dn");
            #
            #$logger->warn(Data::Dumper->Dump([$ref_boot_opts_entry]) );

            # in this case there should be no PXE Link
            my $symlink_rc = removePxeConfigLink(
                entry_dn => $entry_dn,
                config   => $config,
                sc_cache => $sc_cache
            );
        }
    }
    return;
}

sub writeDhcpSettings
{
    my %param = (
        'tmp_fh'            => undef,
        'config'            => undef,
        'boot_opt'          => undef,
        'entry_dn'          => undef,
        'sc_cache'          => undef,
        'dyn_ip'            => 0,
        'netboot_leasetime' => undef,
        'indent_level'      => undef,
        @_,
    );

    my $tmp_fh            = $param{'tmp_fh'};
    my $config            = $param{'config'};
    my $boot_opt          = lc( $param{'boot_opt'} );
    my $entry_dn          = $param{'entry_dn'};
    my $sc_cache          = $param{'sc_cache'};
    my $dyn_ip            = $param{'dyn_ip'};
    my $netboot_leasetime = $param{'netboot_leasetime'};
    my $indent_level;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $indent_level = $param{'indent_level'};
    } else {
        $indent_level = 3;
    }


    #TODO: leasetime should be set per devicetype/bootstate or at least given as parameter
    writeDhcpLeaseTime(
        tmp_fh            => $tmp_fh,
        config            => $config,
        boot_opt          => $boot_opt,
        dyn_ip            => $dyn_ip,
        netboot_leasetime => $netboot_leasetime,
        indent_level      => $indent_level
    );

    writeDhcpNextServer(
        tmp_fh       => $tmp_fh,
        config       => $config,
        boot_opt     => $boot_opt,
        indent_level => $indent_level
    );

    writeDhcpTftpBootOptions(
        tmp_fh            => $tmp_fh,
        config            => $config,
        entry_dn          => $entry_dn,
        boot_opt          => $boot_opt,
        sc_cache          => $sc_cache,
        netboot_leasetime => $netboot_leasetime,
        indent_level      => $indent_level
    );
    return;
}

sub writeDhcpFixed
{
    my %param = (
        'tmp_fh'      => undef,
        'location_dn' => undef,
        'sc_cache'    => undef,
        'boot_opt'    => undef,
        'config'      => undef,
        'zone_name'   => undef,
        @_,
    );

    my $tmp_fh                  = $param{'tmp_fh'};
    my $sc_cache                = $param{'sc_cache'};
    my $location_dn             = $param{'location_dn'};
    my $boot_opt                = $param{'boot_opt'};
    my $config                  = $param{'config'};
    my $zone_name               = $param{'zone_name'};
    my $ip_device_objectclasses =
      [ 'dbkWorkstation', 'dbkPrinter', 'dbkServer' ];

    #print Data::Dumper->Dump([$location_data]); #debug
    my $logger = Log::Log4perl->get_logger();

    # hash used for detecting duplicate IP addresses
    my %ip_check;

    foreach my $objectclass ( @{$ip_device_objectclasses} ) {
        my @ip_devices = $sc_cache->getLocationEntries(
            'base'             => $location_dn,
            'objectclass_name' => $objectclass
        );

        foreach my $dn (@ip_devices) {
            if ( !$sc_cache->hasEntry($dn) ) {
                $logger->warn("[DHCP,LDAP-DATA] no entry found for $dn");
                next;
            }
            my $entry   = $sc_cache->getEntry($dn);
            my $ip_addr = $sc_cache->getIpAddress( entry_dn => $dn );

            if ( $ip_addr =~ /^\s*$/ ) {
                $logger->error( "[DHCP, LDAP-DATA] could not fetch IP for $dn, "
                      . "not generating DHCP fixed address declaration!" );
                next;
            }

            # check for duplicate IP addresses
            if ( $ip_check{$ip_addr} ) {
                $logger->error(
                    "[DHCP,LDAP-DATA] duplicate IP $ip_addr of $dn detected, "
                      . "not generating DHCP fixed address declaration!" );
                next;
            } else {
                $ip_check{$ip_addr}++;
            }

            #my $dns_name      = $sc_cache->getDnsNameForIp($ip_addr);
            my $dns_name =
              ldap_cn2dnsname( $entry->get_value('cn') ) . '.' . $zone_name;
            if ( !$dns_name ) {
                $logger->error(
                    "[DHCP,LDAP-DATA] no dns name could be build for $dn");
            }
            my $mac_address   = $entry->get_value('macAddress');
            my %device_status = $sc_cache->getDeviceStatus($dn);
            if ( !$device_status{'netboot'} ) {
                $device_status{'dbkDoTftpBoot'} = 'LOCALBOOT';
            }

            # FIXME: macaddress and dns name have to be checked before usage

            print $tmp_fh "    host $dns_name {\n";
            print $tmp_fh "      hardware ethernet $mac_address;\n";
            print $tmp_fh "      fixed-address ", $ip_addr, ";\n";

            #TODO: leasetime should be set per devicetype and bootstate
            if ( $device_status{'dbkDoTftpBoot'} ) {
                my $netboot_leasetime =
                  $sc_cache->getNetbootLeasetimeForDn($dn);
                writeDhcpSettings(
                    tmp_fh            => $tmp_fh,
                    config            => $config,
                    boot_opt          => $device_status{'dbkDoTftpBoot'},
                    entry_dn          => $dn,
                    sc_cache          => $sc_cache,
                    netboot_leasetime => $netboot_leasetime,
                    indent_level      => 3
                );
            } else {
                $logger->warn(
                    "[DHCP] could not get dbkDoTftpBoot for $dn from cache");
            }

            print $tmp_fh "    }\n";
        }
    }

    return;
}

my $dynamicIpAllowRegexp = qr{
  ^                      # beginning of the line
  (?:                    # do not capture
       (allow|deny)      # capture ONE match "allow" or "deny"
       -                 # if followed by "-"
       (?!known)         # and not followed by "known" (negative look ahead)
  )?                     # at most one times
  (                      # capture TWO
      (?:                # do not capture
         (?:un)?         # do not capture "un" up too one times
         known|all       # match "known" or "all"
      )                  # end of ignoring capture
      -clients           # match "-clients"
  )                      # end of capture TWO
  $                      # end of line
}x;

sub writeDhcpScopeExtension
{
    my %param = ( fh => undef, entry => undef, indent_level => undef, @_ );
    my $fh    = $param{fh};
    my $entry = $param{entry};
    if ( $entry->exists('dbkDhcpScopeExtension') ) {
        my $space_str;
        if ( $param{'indent_level'} =~ /^\d+$/ ) {
            $space_str = (INDENT_SPACER) x $param{'indent_level'};
        } else {
            $space_str = DEFAULT_SPACER;
        }
        print $fh $space_str . "### BEGIN Scope Extensions ###\n";
        my $dbkDhcpScopeExtensions =
          $entry->get_value( 'dbkDhcpScopeExtension', asref => 1 );
        foreach my $dbkDhcpScopeExtension ( @{$dbkDhcpScopeExtensions} ) {
            print $fh map { $space_str . "$_\n" }
              split( /\n/, $dbkDhcpScopeExtension );
        }
        print $fh $space_str . "### END Scope Extensions ###\n";
        print $fh "\n";
    }
    return;
}

sub writeDhcpSearchList
{
    my %param = (
        fh           => undef,
        sc_cache     => undef,
        config       => undef,
        location     => undef,
        indent_level => undef,
        @_
    );

    my $fh = $param{fh};

    my $space_str;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $space_str = (INDENT_SPACER) x $param{'indent_level'};
    } else {
        $space_str = DEFAULT_SPACER;
    }

    print $fh $space_str .
      "if ((substring (option vendor-class-identifier, 0, 5) = \"Linux\") or (substring (option vendor-class-identifier, 0, 6) = \"dhcpcd\")) {\n";
    print $fh $space_str . "  option domain-name \""
      . getDhcpSearchList(
        location_dn => $param{location},
        sc_cache    => $param{sc_cache},
        config      => $param{config}
      ) . "\";\n";

    # or take the ldap base domain name
    print $fh $space_str . "} else {\n";
    print $fh $space_str . "  option domain-name \""
      . $param{sc_cache}->getLocationZoneName( $param{location} ) . "\";\n";
    print $fh $space_str . "}\n";
    return;
}

sub generateDhcpSharedNetworkConfig
{
    my %param = (
        'sc_cache'             => undef,
        'location'             => undef,
        'shared_network_name'  => undef,
        'shared_locations'     => undef,
        'config'               => undef,
        'authoritative'        => undef,
        'dyn_range'            => undef,
        'nextserver_localboot' => undef,
        @_
    );
    
    my $logger              = Log::Log4perl->get_logger();
    my $changed_location    = $param{location};
    my $shared_network_name = $param{shared_network_name};
    my $shared_locations    = $param{shared_locations};
    my $sc_cache            = $param{sc_cache};
    my $config              = $param{config};
    my $change_detected     = 0;

    my ( $tmp_fh, $tmp_filename ) = tempfile();
    $logger->debug("[DHCP] opened tmp file $tmp_filename successfully");

    print $tmp_fh "  shared-network $shared_network_name {\n\n" ;

    # Write subnet-declarations
    my %location_attrs;
    my $contains_location = 0;
    foreach my $shared_location ( @{$shared_locations} ) {
        
        if ( $sc_cache->checkLocationIgnore( location_dn => $shared_location ) ) {
            $logger->info(
                "[DHCP] location $shared_location matches ignore condition, skipping"
            );
            next;
        }

        my $zone_name = $sc_cache->getLocationZoneName($shared_location);
        if ( !$zone_name ) {
            $logger->error(
                "[DHCP,LDAP-DATA] No zone name for location $shared_location found"
            );
            next;
        }
        
        # write the dn of the current entry as a comment (useful for finding errors in LDAP entries)
        print $tmp_fh "    # dn: ", $shared_location, "\n";

        # Write subnet-declaration 
        my %ref_attrs = generateDhcpSubnet(
            fh                   => $tmp_fh,
            sc_cache             => $sc_cache,
            location             => $shared_location,
            config               => $config,
            authoritative        => $param{authoritative},
            dyn_range            => $param{dyn_range},
            nextserver_localboot => $param{nextserver_localboot},
            indent_level         => 2,
        );
        $location_attrs{$shared_location}{ref_attrs} = \%ref_attrs;
        $location_attrs{$shared_location}{zone_name} = $zone_name;

        $contains_location = 1 if ($changed_location eq $shared_location);
    }

    print $tmp_fh "  }\n\n";
    print $tmp_fh "  group {\n";
    #TODO: leasetime should be set per devicetype/bootstate or at least given as parameter
    writeDhcpLeaseTime(
        tmp_fh       => $tmp_fh,
        config       => $config,
        boot_opt     => 'localboot',
        indent_level => 2
    );

    foreach my $shared_location ( @{$shared_locations} ) {
        # generate fixed-address entries
        writeDhcpFixed(
            tmp_fh      => $tmp_fh,
            location_dn => $shared_location,
            sc_cache    => $sc_cache,
            boot_opt    => $location_attrs{$shared_location}{ref_attrs}{'dbkDoTftpBoot'},
            config      => $config,
            zone_name   => $location_attrs{$shared_location}{zone_name}
        );
    }

    print $tmp_fh "  }\n\n";
    
    close($tmp_fh) or do {
       $logger->warn("[DHCP, LDAP-DATA] Can not close tmp file for shared-network $shared_network_name: $!"); 
    };

    if ( ! $contains_location ) {
        $logger->warn("[DHCP, LDAP-DATA] location $changed_location not in shared network $shared_network_name");
    }

    my $config_filename = getDhcpConfigFileName(
        config              => $config,
        shared_network_name => $shared_network_name
    );

    # compare temporary file with dhcp config file
    if ( File::Compare::compare( $tmp_filename, $config_filename ) != 0 )
    {
        $change_detected = 1;
        if ( !copy( $tmp_filename, $config_filename ) ) {
            $logger->error( "[DHCP] could not copy $tmp_filename to "
                  . $config_filename
                  . ": $!" );
        } else {

            # copy was successful, delete temporary file
            unlink($tmp_filename);
        }
    } else {

        # no change detected, delete temporary file
        unlink($tmp_filename);
    }

    return $change_detected;
    
}

sub generateDhcpLocationConfig
{
    my %param = (
        'sc_cache'             => undef,
        'location'             => undef,
        'config'               => undef,
        'authoritative'        => undef,
        'dyn_range'            => undef,
        'nextserver_localboot' => undef,
        @_
    );

    my $logger           = Log::Log4perl->get_logger();
    my $changed_location = $param{location};
    my $sc_cache         = $param{sc_cache};
    my $config           = $param{config};
    my $change_detected  = 0;

    if ( $sc_cache->checkLocationIgnore( location_dn => $changed_location ) ) {
        $logger->info(
            "[DHCP] location $changed_location matches ignore condition, skipping"
        );
        return;
    }

    my $zone_name = $sc_cache->getLocationZoneName($changed_location);
    if ( !$zone_name ) {
        $logger->error(
            "[DHCP,LDAP-DATA] No zone name for location $changed_location found"
        );
        return;
    }

    my ( $tmp_fh, $tmp_filename ) = tempfile();
    $logger->debug("[DHCP] opened tmp file $tmp_filename successfully");

    # write the dn of the current entry as a comment (useful for finding errors in LDAP entries)
    print $tmp_fh "  # dn: ", $changed_location, "\n";

    # Write subnet-declaration 
    my %ref_attrs = generateDhcpSubnet(
        fh                   => $tmp_fh,
        sc_cache             => $sc_cache,
        location             => $changed_location,
        config               => $config,
        authoritative        => $param{authoritative},
        dyn_range            => $param{dyn_range},
        nextserver_localboot => $param{nextserver_localboot},
        indent_level         => 1,
    );
    
    print $tmp_fh "  group {\n";
    #TODO: leasetime should be set per devicetype/bootstate or at least given as parameter
    writeDhcpLeaseTime(
        tmp_fh       => $tmp_fh,
        config       => $config,
        boot_opt     => $ref_attrs{'dbkDoTftpBoot'},
        indent_level => 2
    );

    # generate fixed-address entries
    writeDhcpFixed(
        tmp_fh      => $tmp_fh,
        location_dn => $changed_location,
        sc_cache    => $sc_cache,
        boot_opt    => $ref_attrs{'dbkDoTftpBoot'},
        config      => $config,
        zone_name   => $zone_name
    );
    print $tmp_fh "  }\n\n";

    close($tmp_fh) or do {
       $logger->warn("[DHCP, LDAP-DATA] Can not close tmp file for location $changed_location: $!"); 
    };

    my $config_filename = getDhcpConfigFileName(
        config => $config,
        zone_name => $zone_name
    );

    # compare temporary file with dhcp config file
    if ( File::Compare::compare( $tmp_filename, $config_filename ) != 0 )
    {
        $change_detected = 1;
        if ( !copy( $tmp_filename, $config_filename ) ) {
            $logger->error( "[DHCP] could not copy $tmp_filename to "
                  . $config_filename
                  . ": $!" );
        } else {

            # copy was successful, delete temporary file
            unlink($tmp_filename);
        }
    } else {

        # no change detected, delete temporary file
        unlink($tmp_filename);
    }

    return $change_detected;

}

sub generateDhcpSubnet
{

    my %param = (
        'fh'                   => undef,
        'sc_cache'             => undef,
        'location'             => undef,
        'config'               => undef,
        'authoritative'        => undef,
        'dyn_range'            => undef,
        'nextserver_localboot' => undef,
        'indent_level'         => undef,
        @_
    );

    my $logger               = Log::Log4perl->get_logger();
    my $fh                   = $param{fh};
    my $sc_cache             = $param{sc_cache};
    my $config               = $param{config};
    my $changed_location     = $param{location};
    my $nextserver_localboot = $param{nextserver_localboot};
    
    $logger->debug("[DHCP] generating subnet for $changed_location");
    my $entry = $sc_cache->getEntry($changed_location);
    if ( !$entry ) {
        $logger->info(
            "[DHCP,LDAP-DATA] entry for $changed_location not found in cache");
        return;
    }

    return if !checkLocationEntry($entry);

    my $network_obj = $sc_cache->getLocationIpObject($changed_location);
    my $network     = $network_obj->network();
    my $netmask     = $network_obj->netmask();

    my $indent_level;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $indent_level = $param{'indent_level'};
    } else {
        $indent_level = 1;
    }
    my $space_str = (INDENT_SPACER) x $indent_level;

    print $fh $space_str . "subnet $network netmask $netmask {\n\n";
    $space_str = (INDENT_SPACER) x ++$indent_level;
    
    # if $not_authoritative is set, write the "not authoritative" flag
    # into the config file to avoid sending DHCPNAKs.
    if ( !$param{authoritative} ) {
        print $fh $space_str . "not authoritative;\n";
    }

    print $fh $space_str . "option broadcast-address ",
      $network_obj->broadcast(), ";\n";
    print $fh $space_str . "option routers ", $entry->get_value('dbkDefaultGw'),
      ";\n\n";

    writeDhcpScopeExtension( fh => $fh, entry => $entry, indent_level => $indent_level );
    writeDhcpSearchList(
        fh           => $fh,
        sc_cache     => $sc_cache,
        config       => $config,
        location     => $changed_location,
        indent_level => $indent_level
    );
    writeDhcpOptions(
        fh           => $fh,
        entry        => $entry,
        sc_cache     => $sc_cache,
        location     => $changed_location,
        indent_level => $indent_level
    );

    # pool does only work if we have a dynamic range
    # a range is only generated if dbkDynamicIp is set to "TRUE"
    # but if the configuration directive LDAP2DHCP_NODYNRANGE is
    # true, never write a range.
    # And dbkDhcpRange must be set in the entry.
    my $dyn_ip          = 0;
    my $dbkDynamicIp    = 'FALSE';
    my $split_dhcp_pool = $config->getWithDefault( 'LDAP2DHCP_SPLIT_DHCP_POOL', 0 );
    my $standard_dhcp_pool_clients;
    my $other_dhcp_pool_clients;
    if ($split_dhcp_pool) {
        $standard_dhcp_pool_clients = 'not known';
        $other_dhcp_pool_clients    = 'known';
    } else {
        $standard_dhcp_pool_clients = 'allow unknown-clients';
    }
    if ( $entry->exists('dbkDynamicIp') ) {
        $dbkDynamicIp = $entry->get_value('dbkDynamicIp');
    } else {
        $logger->warn(
            "[DHCP,LDAP-DATA] dbkDynamicIp not defined for $changed_location. "
              . "Assuming FALSE" );
    }

    if (   $param{dyn_range}
        && $entry->exists('dbkDhcpRange')
        && $dbkDynamicIp eq 'TRUE' )
    {
        $dyn_ip = 1;
        ( $standard_dhcp_pool_clients, $other_dhcp_pool_clients ) =
          calculateDhcpSplitPoolValues(
            entry            => $entry,
            split_pool       => $split_dhcp_pool,
            location         => $changed_location,
            default_standard => $standard_dhcp_pool_clients,
            default_other    => $other_dhcp_pool_clients,
          );
        if ( !$other_dhcp_pool_clients ) {
            $split_dhcp_pool = 0;
        }
    }

    # generate netboot entries (for the whole subnet)
    #            genbootopts( $entry->dn );
    my %ref_attrs = $sc_cache->getReferenceAttributesValues($changed_location);

    if ( $dyn_ip == 1 ) {
        writeDhcpDynRange(
            fh               => $fh,
            split_pool       => $split_dhcp_pool,
            sc_cache         => $sc_cache,
            entry            => $entry,
            location         => $changed_location,
            config           => $config,
            standard_clients => $standard_dhcp_pool_clients,
            other_clients    => $other_dhcp_pool_clients,
            ref_attrs        => \%ref_attrs,
            indent_level     => $indent_level
        );
    }

    # generate pool for known clients if scMobileDevicesIpRange is set
    # or LDAP2DHCP_SPLIT_DHCP_POOL option is set
    writeDhcpMobileRange(
        entry                => $entry,
        config               => $config,
        fh                   => $fh,
        split_pool           => $split_dhcp_pool,
        nextserver_localboot => $nextserver_localboot,
        indent_level         => $indent_level
    );

    # close subnet-declaration
    $space_str = (INDENT_SPACER) x --$indent_level;
    print $fh $space_str . "}\n\n";
    
    return %ref_attrs;
    
}

sub writeDhcpMobileRange
{
    my %param = (
        fh                   => undef,
        entry                => undef,
        config               => undef,
        split_pool           => undef,
        nextserver_localboot => undef,
        indent_level         => undef,
        @_
    );
    my $entry                = $param{entry};
    my $config               = $param{config};
    my $fh                   = $param{fh};
    my $nextserver_localboot = $param{nextserver_localboot};
    my $split_dhcp_pool      = $param{split_pool};
    my $logger               = Log::Log4perl->get_logger();
    if ( $entry->exists('scMobileDevicesIpRange') ) {
        if (!isPrimary($config)) {
            $logger->info("skip generating mobile-range since we are not primary mode");
            return;
        }
        if ($split_dhcp_pool) {
            $logger->warn(
                "[LDAP] split dhcp pool is set globally, so this scMovileDevicesIpRange is not really useful and will not be generated"
            );
            return;
        }

        my @range =
          splitDhcpRange( $entry->get_value('scMobileDevicesIpRange') );

        my $space_str;
        if ( $param{'indent_level'} =~ /^\d+$/ ) {
            $space_str = (INDENT_SPACER) x $param{'indent_level'};
        } else {
            $space_str = DEFAULT_SPACER;
        }

        print $fh $space_str . "pool {\n";
        print $fh $space_str . INDENT_SPACER . "deny unknown-clients;\n";
        print $fh $space_str . INDENT_SPACER . "default-lease-time ";
        print $fh $config->get( 'LDAP2DHCP_LOCALBOOTLEASETIME' ), ";\n";
        print $fh $space_str . INDENT_SPACER . "max-lease-time ";
        print $fh $config->get( 'LDAP2DHCP_LOCALBOOTLEASETIME' ), ";\n";
        print $fh $space_str . INDENT_SPACER . "next-server ";
        print $fh $nextserver_localboot, ";\n";
        print $fh $space_str . INDENT_SPACER . 'filename "LOCALBOOT";', "\n";
        print $fh $space_str . INDENT_SPACER . "range ", $range[0], ONE_SPACE, $range[1], ";\n";
        print $fh $space_str . "}\n\n";
    }
    return;
}

sub writeDhcpDynRange
{
    my %param = (
        fh               => undef,
        split_pool       => undef,
        sc_cache         => undef,
        entry            => undef,
        location         => undef,
        config           => undef,
        standard_clients => undef,
        other_clients    => undef,
        ref_attrs        => undef,
        indent_level     => undef,
        @_
    );
    my $fh                         = $param{fh};
    my $sc_cache                   = $param{sc_cache};
    my $entry                      = $param{entry};
    my $changed_location           = $param{location};
    my $config                     = $param{config};
    my $split_dhcp_pool            = $param{split_pool};
    my $standard_dhcp_pool_clients = $param{standard_clients};
    my $other_dhcp_pool_clients    = $param{other_clients};
    my %ref_attrs                  = %{ $param{ref_attrs} };
    my $logger                     = Log::Log4perl->get_logger();

    my $standard_pool_boot_options;
    my $other_pool_boot_options;

    my $indent_level;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $indent_level = $param{'indent_level'};
    } else {
        $indent_level = 2;
    }
    my $space_str = (INDENT_SPACER) x $indent_level;

    print $fh $space_str . "pool {\n";
    $space_str = (INDENT_SPACER) x ++$indent_level;
    if ($split_dhcp_pool) {

        # Write settings for standard dynamic clients (value of $dbkDynamicIpAllow)
        my $standard_boot_entry =
          $sc_cache->getEntry( $ref_attrs{'scRefBootOptionsDn'} );
        if ( !ref $standard_boot_entry ) {
            $logger->warn(
                "[LDAP, DHCP] no cache entry found for $ref_attrs{'scRefBootOptionsDn'} "
                  . "using localboot" );
        }
        my $boot_from = lc( $ref_attrs{'dbkDoTftpBoot'} );
        $standard_pool_boot_options = calculateDhcpStandardBootOptions(
            boot_from  => $boot_from,
            boot_entry => $standard_boot_entry
        );
        print $fh $space_str . "if $standard_dhcp_pool_clients {\n";
        $space_str = (INDENT_SPACER) x ++$indent_level;

        #TODO: leasetime should be set per devicetype/bootstate or at least given as parameter
        writeDhcpLeaseTime(
            tmp_fh       => $fh,
            config       => $config,
            boot_opt     => $boot_from,
            dyn_ip       => 1,
            indent_level => $indent_level
        );
        writeDhcpNextServer(
            tmp_fh    => $fh,
            config    => $config,
            boot_opt  => $boot_from,
            indent_level => $indent_level
        );
        print $fh $space_str . "$standard_pool_boot_options\n";
        $space_str = (INDENT_SPACER) x --$indent_level;
        print $fh $space_str . "}\n";

        # Write settings for other dynamic clients (opposite value of $dbkDynamicIpAllow)
        $other_pool_boot_options = calculateDhcpOtherBootOptions(
            other_pool_ref => lc(
                $config->getWithDefault( 'LDAP2DHCP_OTHER_DHCP_POOL_BOOTREF', 'standard' )
            ),
            sc_cache                 => $sc_cache,
            location_boot_state_href => \%ref_attrs,
            boot_from                => $boot_from,
        );
        print $fh $space_str . "if $other_dhcp_pool_clients {\n";
        $space_str = (INDENT_SPACER) x ++$indent_level;

        #TODO: leasetime should be set per devicetype/bootstate or at least given as parameter
        writeDhcpLeaseTime(
            tmp_fh       => $fh,
            config       => $config,
            boot_opt     => $boot_from,
            dyn_ip       => 1,
            indent_level => $indent_level
        );
        writeDhcpNextServer(
            tmp_fh       => $fh,
            config       => $config,
            boot_opt     => $boot_from,
            indent_level => $indent_level
        );
        print $fh $space_str . "$other_pool_boot_options\n";
        $space_str = (INDENT_SPACER) x --$indent_level;
        print $fh $space_str . "}\n";

    } else {
        print $fh $space_str . "$standard_dhcp_pool_clients;\n";

        if ( $ref_attrs{'dbkDoTftpBoot'} ) {
            writeDhcpSettings(
                tmp_fh    => $fh,
                config    => $config,
                boot_opt  => $ref_attrs{'dbkDoTftpBoot'},
                entry_dn  => $changed_location,
                sc_cache  => $sc_cache,
                dyn_ip    => 1,
                indent_level => $indent_level
            );
        } else {
            $logger->warn(
                "[DHCP] could not get dbkDoTftpBoot for $changed_location from cache"
            );
        }
    }

    my @dhcp_range;
    if ( $entry->exists('dbkDhcpRange') ) {
        my $tmprange = $entry->get_value('dbkDhcpRange');
        if ( defined($tmprange) ) {
            $logger->debug(
                "[DHCP] dbkDhcpRange for $changed_location is $tmprange");
            @dhcp_range = splitDhcpRange($tmprange);
            print $fh $space_str . "range ", $dhcp_range[0], ONE_SPACE,
              $dhcp_range[1], ";\n";
        }
    } else {
        $logger->warn(
            "[LDAP-DATA, DHCP] no dhcp range for dynamic ip's configured in $changed_location"
        );
    }
    $space_str = (INDENT_SPACER) x --$indent_level;
    print $fh $space_str . "}\n\n";

    return;
}

sub calculateDhcpOtherBootOptions
{
    my %params = (
        other_pool_ref           => undef,
        sc_cache                 => undef,
        location_boot_state_href => undef,
        boot_from                => undef,
        @_
    );
    my $sc_cache                = $params{sc_cache};
    my $other_dhcp_pool_bootref = $params{other_pool_ref};
    my %location_boot_state     = %{ $params{location_boot_state_href} };
    my $boot_from               = $params{boot_from};
    my $logger                  = Log::Log4perl->get_logger();

    if ( $other_dhcp_pool_bootref ne 'localboot' ) {
        my $boot_ref_entry;
        if ( $other_dhcp_pool_bootref ne 'standard' ) {
            $boot_ref_entry = $sc_cache->getEntry($other_dhcp_pool_bootref);
            if ( !ref $boot_ref_entry ) {
                # FIXME: the warn message seems to indicate another value for $boot_ref_entry than standard
                $logger->warn(
                    "[CONFIG, DHCP] no cache entry found for $other_dhcp_pool_bootref "
                      . "using standard $location_boot_state{'scRefBootOptionsDn'} of Location"
                );
                $boot_ref_entry = 'standard';
            }
        }
        if ( ref $boot_ref_entry ) {
            if ( $boot_from eq 'localboot' ) {
                return $boot_ref_entry->get_value('dbkDhcpOptionsLocal')
                  || "filename \"LOCALBOOT\";";
            } else {
                return $boot_ref_entry->get_value('dbkDhcpOptionsRemote')
                  || "filename \"LOCALBOOT\";";
            }
        }
    }
    return "filename \"LOCALBOOT\";";
}

sub calculateDhcpStandardBootOptions
{
    my %params = (
        boot_from  => undef,
        boot_entry => undef,
        @_
    );
    my $boot_from           = $params{boot_from};
    my $standard_boot_entry = $params{boot_entry};
    if ( $boot_from eq 'localboot' ) {
        return $standard_boot_entry->get_value('dbkDhcpOptionsLocal')
          || "filename \"LOCALBOOT\";";
    }
    return $standard_boot_entry->get_value('dbkDhcpOptionsRemote')
      || "filename \"LOCALBOOT\";";
}

sub calculateDhcpSplitPoolValues
{
    my %params = (
        entry            => undef,
        split_pool       => undef,
        location         => undef,
        default_standard => undef,
        default_other    => undef,
        @_
    );
    my $entry            = $params{entry};
    my $split_dhcp_pool  = $params{split_pool};
    my $changed_location = $params{location};
    my $logger = Log::Log4perl->get_logger();
    if ( $entry->exists('dbkDynamicIpAllow') ) {
        my $dbkDynamicIpAllow = $entry->get_value('dbkDynamicIpAllow');

        # Following entries are recognized from isc DHCP Server V3.0.1rc13
        # (tested under SLES 9 dhcpcd-1.3.22pl4-193.8):
        # ALLOW-UNKNOWN-CLIENTS -> allow unknown-clients
        # DENY-UNKNOWN-CLIENTS -> deny unknown-clients
        # ALLOW-ALL-CLIENTS -> allow all clients
        # DENY-ALL-CLIENTS -> deny all clients
        # In these cases the pool will not be splitted!
        #
        # For a split pool (seperate boot options for known and unknown clients),
        # the entries UNKNOWN-CLIENTS and KNOWN-CLIENTS are recognized
        #
        # If dbkDynamicIpAllow is not set, it is assumed UNKNOWN-CLIENTS!
        if ( lc($dbkDynamicIpAllow) =~ m/$dynamicIpAllowRegexp/x ) {
            my ( $permit_stmt, $client_stmt ) = ( $1, $2 );
            $permit_stmt ||= "allow";
            if (
                   $split_dhcp_pool
                && !$1
                && (   $client_stmt eq 'unknown-clients'
                    || $client_stmt eq 'known-clients' )
              )
            {
                if ( $client_stmt eq 'unknown-clients' ) {
                    return ( 'not known', 'known' );
                } else {
                    return ( 'known', 'not known' );
                }
            } else {
                if ($split_dhcp_pool) {
                    $logger->info(
                        "[DHCP,LDAP-DATA] LDAP2DNS_SPLIT_DHCP_POOL is set and "
                          . "dbkDynamicIpAllow="
                          . $dbkDynamicIpAllow
                          . " for $changed_location, "
                          . "pool will not be splitted" );
                }
                my $standard_dhcp_pool_clients =
                  $permit_stmt . ONE_SPACE
                  . (
                    $client_stmt =~ m/\s*all-clients\s*/
                    ? "all clients"
                    : $client_stmt
                  );
                return ( $standard_dhcp_pool_clients, undef );
            }
        } else {
            $logger->warn( " [ DHCP, LDAP-DATA ] unknown value $dbkDynamicIpAllow "
                  . "in attribute dbkDynamicIpAllow in $changed_location, " 
                  . "assuming UNKNOWN-CLIENTS ");
        }
    } else {
        $logger->info( " [ DHCP, LDAP-DATA ] dbkDynamicIpAllow not defined for "
              . $changed_location . ", "
              . " assuming UNKNOWN-CLIENTS " );
    }

    return ( $params{default_standard}, $params{default_other} );
}

sub writeDhcpOptions
{
    my %param =
      ( fh => undef, sc_cache => undef, location => undef, entry => undef, indent_level => undef, @_ );
    my $fh    = $param{fh};
    my $entry = $param{entry};

    my $space_str;
    if ( $param{'indent_level'} =~ /^\d+$/ ) {
        $space_str = (INDENT_SPACER) x $param{'indent_level'};
    } else {
        $space_str = DEFAULT_SPACER;
    }

    print $fh $space_str . "option subnet-mask ",
      $entry->get_value('ipNetmaskNumber'), ";\n";
    my @dhcp_opts = getDhcpOptions(
        location_dn => $param{location},
        sc_cache    => $param{sc_cache}
    );
    if (@dhcp_opts) {
        foreach my $dhcp_option (@dhcp_opts) {
            print $fh $space_str . "option " . $dhcp_option . ";\n";
        }
    }
    print $fh "\n";
    return;
}

sub generateDhcpConfigFiles
{
    my %param = (
        'sc_cache'          => undef,
        'config'            => undef,
        'changed_locations' => undef,
        @_,
    );
    my $sc_cache               = $param{'sc_cache'};
    my $config                 = $param{'config'};
    my $changed_locations      = $param{'changed_locations'};
    my $nextserver_localboot   = 'localboot';
    my $change_detected = 0;

    my $logger = Log::Log4perl->get_logger();

    # Flag for overriding Attribute dbkDynamicIp
    # if the config directive LDAP2DHCP_NODYNRANGE
    # is set to 1, a dynamic DHCP range will never
    # be used.
    my $no_dyn_range = $config->getWithDefault( 'LDAP2DHCP_NODYNRANGE', 0, 'DHCP' );

    # Parameter for setting the DHCP parameter "not authoritative"
    my $not_authoritative =
      $config->getWithDefault( 'LDAP2DHCP_NOTAUTHORITATIVE', 0, 'DHCP' );

    my ( $tmp_fh_main, $tmp_filename_main ) = tempfile();
    $logger->debug("[DHCP] opened tmp file $tmp_filename_main successfully");
    if ( !writeDhcpHeader( $tmp_fh_main, $config->get( 'LDAP2DHCP_TEMPLATEFILE' ) ) )
    {
        close($tmp_fh_main) or do {
            $logger->warn("[DHCP] Can not close $tmp_filename_main: $!");
        };
        return;
    }

    my @dhcp_locations;
    my %shared_networks;
    foreach my $location_dn ($sc_cache->getAllLocationDns()) {
        next
          if ( $sc_cache->checkLocationIgnore( location_dn => $location_dn )
            || !$sc_cache->hasEntry($location_dn) );
        my $location_entry = $sc_cache->getEntry($location_dn);
        my $dhcp_location = $location_dn;
        push @dhcp_locations, $dhcp_location;
        if ( $location_entry->exists('dbkSharedNetwork') ) {
            my $shared_network_name = $location_entry->get_value('dbkSharedNetwork');
            push @{$shared_networks{$shared_network_name}} , $dhcp_location;
        }
    }

    foreach my $changed_location ( @{$changed_locations} ) {
        my $local_change_detected;
        my $location_entry = $sc_cache->getEntry($changed_location);
        if ( $location_entry->exists('dbkSharedNetwork') ) {
            my $shared_network_name = $location_entry->get_value('dbkSharedNetwork');
            $local_change_detected = generateDhcpSharedNetworkConfig(
                sc_cache             => $sc_cache,
                location             => $changed_location,
                shared_network_name  => $shared_network_name,
                shared_locations     => $shared_networks{$shared_network_name},
                config               => $config,
                authoritative        => !$not_authoritative,
                dyn_range            => !$no_dyn_range,
                nextserver_localboot => $nextserver_localboot,
            );
            $change_detected ||= $local_change_detected;
        } else {
            $local_change_detected = generateDhcpLocationConfig(
                sc_cache             => $sc_cache,
                location             => $changed_location,
                config               => $config,
                authoritative        => !$not_authoritative,
                dyn_range            => !$no_dyn_range,
                nextserver_localboot => $nextserver_localboot,
            );
            $change_detected ||= $local_change_detected;
        }
    }
    
    my %shared_network_include;
    foreach my $dhcp_location (@dhcp_locations) {
        my $config_filename;
        my $location_entry = $sc_cache->getEntry($dhcp_location);
        if ( $location_entry->exists('dbkSharedNetwork') ) {
            my $shared_network_name = $location_entry->get_value('dbkSharedNetwork');
            next if ( exists($shared_network_include{$shared_network_name} ));
            $config_filename = getDhcpConfigFileName(
                config => $config,
                shared_network_name => $shared_network_name
            );
            $shared_network_include{$shared_network_name} = 1;
        } else {
            $config_filename = getDhcpConfigFileName(
                config => $config,
                zone_name => $sc_cache->getLocationZoneName($dhcp_location)
            );
        }
        print $tmp_fh_main 'include "' . $config_filename . '";', "\n";
    }
    close($tmp_fh_main) or do {
        $logger->warn("[DHCP] Can not close $tmp_filename_main: $!");
    };
    return ( $tmp_filename_main, $change_detected );
}

sub getDhcpConfigFileName
{
    my %param = (
        'config'              => undef,
        'zone_name'           => undef,
        'shared_network_name' => undef,
        @_
    );
    my $config = $param{'config'};
    my $logger = Log::Log4perl->get_logger();
    if (! $param{'zone_name'} && ! $param{'shared_network_name'} )
    {
        $logger->error("[DHCP] It was neither the parameter zone_name nor the parameter shared_network_name specified");
        return;
    }
    my $dhcp_cfg_name = ($param{'zone_name'}) ? $param{'zone_name'} : $param{'shared_network_name'};
    return $config->get( 'LDAP2DHCP_DHCPCONFPATH' ) . "/$global_current_dhcp_subdir/" . $dhcp_cfg_name . '.conf';
}

# TODO: This method should be in ScCache or even better in a location object.
sub checkLocationEntry
{
    my ($entry)         = @_;
    my $dn              = $entry->dn;
    my $logger          = Log::Log4perl->get_logger();
    my %test_attributes = (
        'dbkDefaultGw'    => { level => 'error' },
        'ipNetmaskNumber' => { level => 'error' },
        'ipNetworkNumber' =>
          { level => 'warn', description => 'Probably VPN location.' }
    );
    my $has_all_test_attributes = 1;
    foreach my $test_attribute ( keys %test_attributes ) {
        if ( !$entry->exists('ipNetworkNumber')
            || ( $entry->get_value($test_attribute) =~ /^\s*$/ ) )
        {
            my %info    = $test_attributes{$test_attribute};
            my $message =
              "[DHCP, LDAP-DATA] no $test_attribute defined for $dn. "
              . ( $info{description} || '[no description]' );
            if ( $info{level} eq 'error' ) {
                $logger->error($message);
            } else {
                $logger->warn($message);
            }
            $has_all_test_attributes = 0;
        }
    }
    return $has_all_test_attributes;
}

sub checkDhcpConfSyntax
{
    my ($dhcpd_conf_tmp_filename) = @_;

    my $return_code        = 0;
    my $debug_filename     = '/tmp/dhcpconferr.out';
    my $dhcpd_conf_invalid = '/tmp/dhcpd.conf.with-errors';

    my $logger = Log::Log4perl->get_logger();

    my $dhcpd_check_exit_code =
      system(
        "/usr/sbin/dhcpd -t -cf $dhcpd_conf_tmp_filename 2> $debug_filename") /
      256;
    if ( $dhcpd_check_exit_code == 0 ) {

        # syntax is ok, return true
        return 1;
    }

    copy( $dhcpd_conf_tmp_filename, $dhcpd_conf_invalid );

    # log syntax check error message
    $logger->fatal("[DHCP] dhcpd syntax check failed, keeping old config");
    $logger->fatal("[DHCP] see $debug_filename for details");
    $logger->fatal("[DHCP] the file $dhcpd_conf_invalid contains errors");
    return 0;
}

=head2 callNotifyOfModification

notifies sc_cache of changes for a list of dn

=over 4

=item Parameter

=item Return

=back

=cut

sub callNotifyOfModification
{
    my ($aref_list_of_dns) = @_;

    my %result =
      ( 'changed_dns' =>
          $global_sc_cache->notifyOfModification($aref_list_of_dns) );

    return %result;
}

sub generateDhcpdConfig
{
    my %param = (
        'sc_cache' => undef,
        'config'   => undef,
        @_,
    );
    my $sc_cache          = $param{'sc_cache'};
    my $config            = $param{'config'};
    my $changed_locations = $param{'changed_locations'};

    my $logger = Log::Log4perl->get_logger();

    my ( $dhcpd_conf_tmp_filename, $change_detected ) = generateDhcpConfigFiles(
        sc_cache          => $sc_cache,
        config            => $config,
        changed_locations => $changed_locations
    );

    my $dhcpd_conf_filename = $config->get( 'LDAP2DHCP_DHCPCONFFILE' );
    my $new_config_command  = $config->get( 'LDAP2DHCP_NEW_CONFIG_CMD' );

    if ( ( -f $dhcpd_conf_filename ) && ( !$change_detected ) ) {
        # no changes needed, delete temporary file
        unlink($dhcpd_conf_tmp_filename);
        # no change in any subnet
        return 1;
    }

    # before activating the new config, do a syntax check
    if ( !checkDhcpConfSyntax($dhcpd_conf_tmp_filename) ) {
        # we will keep the temp file $dhcpd_conf_tmp_filename
        return 0;
    }

    # files differ and sytax is ok
    # make copy of last config
    if ( !copy( $dhcpd_conf_filename, $dhcpd_conf_filename . '.old' ) ) {
        $logger->error( "[DHCP] copying $dhcpd_conf_filename to "
              . "$dhcpd_conf_filename.old failed: $!" );
    }

    # copy temporary file to config file
    if ( !copy( $dhcpd_conf_tmp_filename, $dhcpd_conf_filename ) ) {
        $logger->error( "[DHCP] copying $dhcpd_conf_tmp_filename to "
              . "$dhcpd_conf_filename failed: $!" );
        # we will keep the temp file $dhcpd_conf_tmp_filename
        return 2;
    } else {

        # copy was successful, delete temporary file
        unlink($dhcpd_conf_tmp_filename);

        $global_current_dhcp_subdir =
          switchCurrentSubDir( $config, $global_current_dhcp_subdir );

    }

    # activate the new config
    system($new_config_command);

    return 1;
}

sub processAddedLocations
{
    my %params = (
        sc_cache => undef,
        @_
    );

    my $sc_cache = $params{sc_cache};
    my @added_zones = ();
    my $logger = Log::Log4perl->get_logger();

    foreach
      my $added_location_dn ( keys %{ $sc_cache->{'added_location'} } )
    {
        my $added_location_zone_name =
          $sc_cache->getZoneNameForLocation($added_location_dn);
        $logger->info(
            "[DNS] added location $added_location_dn forces reload of dnszone $added_location_zone_name"
        );
        push @added_zones, $added_location_zone_name;
        my $href_rev_zones = $sc_cache->{'rev_zones'};
        foreach my $added_rev_zone ( keys %{$href_rev_zones} ) {
            if (    exists $href_rev_zones->{$added_rev_zone}->{'locations'}
                and exists $href_rev_zones->{$added_rev_zone}->{'locations'}
                ->{$added_location_dn} )
            {
                $logger->info(
                    "[DNS] added location $added_location_dn forces reload of rev dns zone $added_rev_zone"
                );
                push @added_zones, $added_rev_zone;
            }
        }
    }
    $sc_cache->{'added_location'} = {};

    return @added_zones;
}

sub processRemovedLocations
{
    my %params = (
        sc_cache => undef,
        @_
    );
    my @removed_zones = ();
    my $sc_cache = $params{sc_cache};
    my $logger = Log::Log4perl->get_logger();

    if ( keys %{ $sc_cache->{'removed_location'} } ) {

        #this should trick dns_reload to reload a zone with name ""
        $logger->info("[DNS] removed_location should force global dns reload");
        push @removed_zones, EMPTY_STR;
    }
    $sc_cache->{'removed_location'} = {};

    return @removed_zones;
}

sub processChangedLocations
{
    my %params = (
        sc_cache          => undef,
        changed_locations => undef,
        config            => undef,
        @_
    );

    my $sc_cache               = $params{sc_cache};
    my $changed_locations_aref = $params{changed_locations};
    my $config                 = $params{config};
    my $logger                 = Log::Log4perl->get_logger();

    if ( @{$changed_locations_aref} ) {
        my @changed_zones = ();
        if ( isPrimary($config) ) {
            @changed_zones = generateLocationZones(
                changed_locations => $changed_locations_aref,
                ldap_cache        => $sc_cache,
                config            => $config
            );
        }

        if ( $logger->is_debug() ) {
            $logger->debug("changed zones:");
            foreach my $changed_zone (@changed_zones) {
                $logger->debug($changed_zone);
            }
        }
        generateZoneConfigFile(
            ldap_cache => $sc_cache,
            config     => $config
        );

        # check the zone config file
        if ( $config->isEnabled( 'LDAP2DNS_NAMED_CHECKCONF_ACTIVE' ) ) {
            $logger->info("[DNS] calling checkZoneConfigFile");
            checkZoneConfigFile( config => $config );
        }

        push @changed_zones, processAddedLocations( sc_cache => $sc_cache );
        push @changed_zones, processRemovedLocations( sc_cache => $sc_cache );

        if ( isPrimary($config) ) {
            reloadChangedDnsZones(
                changed_zones_aref => \@changed_zones,
                config             => $config
            );
        } else {
            reloadSlaveDns( config => $config );
        }

        generateDhcpdConfig(
            sc_cache          => $sc_cache,
            config            => $config,
            changed_locations => $changed_locations_aref
        );

    }

    return;
}

=head2 mainLoop()

This is the main loop for the generation of DNS and DHCP configuration.


=over 4

=item Parameter

 none

=item Return

This function does never return.

=back

=cut

sub mainLoop
{
    my $config = SmartClient::Config->new(
        "/etc/smartclient/base",      "/etc/smartclient/ldap2dns",
        "/etc/smartclient/controld",  "/etc/smartclient/leases2ldap",
        "/etc/smartclient/ldap2dhcp", "/etc/smartclient/ldap2dnsdhcpd"
    );

    my @changed_locations;    # list of changed locations

    Log::Log4perl::init( $config->get( 'LDAP2DNSDHCPD_LOG4PERL_CONF' ) );
    my $logger = Log::Log4perl->get_logger();

    $global_current_dhcp_subdir = readCurrentDhcpSubDirState($config);

    $logger->info( basename($0) . " starting mainLoop" );

    #$logger->debug( basename($0) . Data::Dumper->Dump( [$config] ) );

    # check if configuration attributes are defined
    $config->require(
        'ALL_LDAPSERVER_TYPE',
        'ALL_LDAP_FILTER_GET_ALL',
        'CONTROLD_XML_RPC_TIMEOUT',
        'LDAP2DNS_ZONEFILEPATH',
        'LDAP2DNS_ZONETEMPLATE',
        'LDAP2DNS_ZONECONFNAME',
        'LDAP2DNS_DNSMASTERHOST',
        'LDAP2DNS_LOCATION_DNS_DOMAIN_ATTR',
        'LDAP2DNSDHCPD_HA_MODE',
        'LDAP2DNS_DEVICE_STATUS_MMAP_FILE',
        'LDAP2DHCP_APPEND_SEARCHLIST',
        'LDAP2DHCP_TEMPLATEFILE',
        'LDAP2DHCP_LOCALBOOTLEASETIME',
        'LDAP2DHCP_DHCPCONFFILE',
        'LDAP2DHCP_DHCPCONFPATH',
        'LEASES2LDAP_MMAP_FILE_DISLOCATED_DEVICES',
        'LDAP2DNSDHCPD_LDAP_ADMIN_DN',
        'LDAP2DNSDHCPD_LDAP_ADMIN_PW',
    );

    # try until we get LDAP connection
    while ( !$global_sc_ldap ) {
        $global_sc_ldap = SmartClient::LDAP->new(
            'CONFIG'     => $config,
            'LDAPBINDDN' => $config->get( 'LDAP2DNSDHCPD_LDAP_ADMIN_DN' ),
            'LDAPBINDPW' => $config->get( 'LDAP2DNSDHCPD_LDAP_ADMIN_PW' ),
            'NEEDCAP'    => 'ri'
        );

        # TODO: define sleep time in config
        sleep 10 if ( !$global_sc_ldap );
    }

    my $function_queue_page_size =
      $config->getWithDefault( 'SHARED_FUNCTIONS_MMAP_PAGE_SIZE', '512k' );
    my $function_queue_num_pages =
      $config->getWithDefault( 'SHARED_FUNCTIONS_MMAP_NUM_PAGES', '313' );

    $logger->info( basename($0) . " initialising shared function queue" );
    $logger->info( "HA MODE: " . $config->get( 'LDAP2DNSDHCPD_HA_MODE' ) );
    my $function_queue = SmartClient::SharedFunction->new(
        'shm_id_queue'          => 'l2dq',
        'mmap_result_file'      => '/var/cache/smartclient/shared_function_results',
        'init_mmap_result_file' => 1,
        'mmap_page_size'        => $function_queue_page_size,
        'mmap_num_pages'        => $function_queue_num_pages,
    );

    $global_sc_cache = SmartClient::ScCache->new(
        'config' => $config,
        ## mmap status filename is a seperate parameter,
        ## to prevent problems if multiple applications uses ScCache
        ## if not specified, no mmap device status file is used
        'mmap_device_status_filename' =>
          $config->get( 'LDAP2DNS_DEVICE_STATUS_MMAP_FILE' )
    );

    my $ldapFilterGetAll = $config->get( 'ALL_LDAP_FILTER_GET_ALL' );

    $global_sc_cache->generateEntryCache(
        sc_ldap => $global_sc_ldap,
        filter  => $ldapFilterGetAll
    );

    # add these objectclasses explicitly to the index
    foreach my $oc_name (
        'dbkWorkstation', 'dbkServer',
        'dbkService',     'dbkPrinter',
        'dbkDeviceTypeStandards'
      )
    {
        $global_sc_cache->addObjectClassIndex($oc_name);
    }
    $logger->info("assign networks to locations");
    $global_sc_cache->assignNetworksToLocations();
    $logger->info("assign ip devices");
    $global_sc_cache->assignIpDevices();
    if ( isPrimary($config) ) {
        $logger->info("generate serials for locations");
        $global_sc_cache->generateSerialsForLocations();
    }

    $logger->info("create leases2ldap");
    $global_leases2ldap = SmartClient::Leases2ldap->new(
        'config'   => $config,
        'sc_ldap'  => $global_sc_ldap,
        'sc_cache' => $global_sc_cache
    );

    #my $device_status_file = createDeviceStatusFile( 'config' => $config );

    # when starting all locations are considered as changed
    @changed_locations = $global_sc_cache->getAllLocationDns();

    #print Data::Dumper->Dump([@changed_locations]);

    # get the XMLRPC-Connection to controld of secondary server
    my $controld_secondary = $global_sc_cache->getXmlRpcConnection();
    $logger->info( "could not create XMLRPC-Connection to controld of secondary"
          . " server (if ALL_CONTROLD_SECONDARY_URL not configured"
          . " this might be OK)" )
      if ( !$controld_secondary );

    $logger->info("starting while(1) in mainLoop");
    while (1) {
        my $loop_start_time = [gettimeofday];

        processChangedLocations(
            sc_cache          => $global_sc_cache,
            changed_locations => \@changed_locations,
            config => $config,
        );

        $global_sc_cache->getStatusCache()->commitUpdates();

        # clear the leases from added/moved workstations
        $logger->info("clear leases");
        $global_leases2ldap->clearLeases();

        # try move/add workstations in ldap
        $logger->info("check for dislocated devices");
        $global_leases2ldap->refreshDislocatedDevicesFromLeases();
        $logger->info("perform actions");
        $global_leases2ldap->performActions();

        # cache and config files should be in harmony
        # process pending shared functions
        my $duration = 5;
        my $rounds   = 12;
        $logger->info(
            "processing shared function queue for the next ${rounds}x${duration}s"
        );
        $function_queue->run( $duration, $rounds );

        # - refresh cache
        # - derive list of changed locations
        my @dn_changed = ();
        if ( isPrimary($config) || $global_create_config_trigger ) {
            $logger->info("refresh entry cache");
            @dn_changed =
              $global_sc_cache->refreshEntryCache( sc_ldap => $global_sc_ldap );

            # reset create config trigger if it was set
            $global_create_config_trigger = 0 if $global_create_config_trigger;
        }

        #my %locations_changed;
        @changed_locations = ();
        if (@dn_changed) {
            $logger->info("changed DNs:");
            foreach my $changed_dn (@dn_changed) {
                $logger->info($changed_dn);
            }
            @changed_locations =
              $global_sc_cache->getAssignedLocations(@dn_changed);
            $logger->info("changed locations:");
            foreach my $changed_location (@changed_locations) {
                $logger->info($changed_location);
            }
            $global_sc_cache->assignNetworksToLocations(
                refreshed_locations_aref => \@changed_locations );
            $global_sc_cache->assignIpDevices(
                refreshed_locations_aref => \@changed_locations );

            if ( isPrimary($config) ) {
                $global_sc_cache->generateSerialsForLocations(
                    refreshed_locations_aref => \@changed_locations );
                if ($controld_secondary) {
                    my $resp = createSecondaryConfig(
                        rpc_xml_daemon => $controld_secondary,
                        timeout        => 0
                    );
                }
            }
        }

        my $loop_duration = tv_interval( $loop_start_time, [gettimeofday] );
        if ( $loop_duration < $config->get( 'CONTROLD_XML_RPC_TIMEOUT' ) ) {
            $logger->debug("duration: $loop_duration");
        } else {
            $logger->warn(
                    "duration: $loop_duration (> CONTROLD_XML_RPC_TIMEOUT ("
                  . $config->get( 'CONTROLD_XML_RPC_TIMEOUT' )
                  . "))" );
        }
    }

    # we never get here, and return nothing
    return;
}

###########################################################
# functions for SmartClient::SharedFunction

=head1 Shared Functions for SmartClient::SharedFunction

=cut

=head2 callScheduleUpdateWorkstations(%)

    called via shared function queue

=over 4

=item Parameter

    struct

=item Return

    

=back

=cut

sub callGetScheduleUpdateWorkstations
{
    my %params = (
        'location_dn'   => undef,
        'refimage_dn'   => undef,
        'devicetype_dn' => undef,
        @_
    );

    my $logger = Log::Log4perl->get_logger();
    $logger->debug( Dumper(%params) );

    my $href_result = $global_sc_cache->getScheduleUpdateWorkstations(%params);

    $logger->debug( "result: " . Dumper($href_result) );

    if ( ref $href_result eq "HASH" ) {

        # SharedFunction always expect a hash as return value
        return %{$href_result};
    }
    return;
}

=head2 callGetDeviceStatus

This is called by sc_ldap2dnsdhcp-statusd.pl for the
XMLRPC method getDeviceStatus

=over 4

=item Parameter

 device DN

=item Return

returns a hash for being processed into a XMLRPC struct.

=back

=cut

sub callGetDeviceStatus( $ )
{
    my ($dn)   = @_;

    my %result = $global_sc_cache->getDeviceStatus($dn);

    return %result;
}

=head2 callDumpEntry

called via sharef function queue
dumps an internal hash entry to the standard logfile with level info

=over 4

=item Parameter

dn of the entry in the entrycache

=item Return

hash containing

dn => message, if dump could be completed

=back

=cut

sub callDumpEntry
{
    my ($dn) = @_;

    my $logger     = Log::Log4perl->get_logger();
    my $href_entry = $global_sc_cache->getInternalEntryHash($dn);
    if ( ref $href_entry ) {
        if ( $logger->is_info() ) {
            $logger->info( Dumper($href_entry) );
        }

        #shared function queue expect a hash as return value
        return %{$href_entry};
    } else {
        return "$dn NOT OK. No entry could be found";
    }
    return;
}

=head2 callChangeDeviceType

=over 4

=item Parameter

=item Return

=back

=cut

sub callChangeDeviceType
{
    my %params = (
        'dn'             => undef,
        'device_type_dn' => undef,
        @_
    );

    my $logger = Log::Log4perl->get_logger();

    my $dn             = $params{'dn'};
    my $device_type_dn = $params{'device_type_dn'};

    if ( !$dn || !$device_type_dn ) {
        my $param_error =
            '[XMLRPC] no entry or new devicetype was given dn:'
          . ( $dn || "undef" )
          . ' devicetype: '
          . ( $device_type_dn || 'undef' );
        $logger->warn($param_error);
        return ( 'code' => 9, 'message' => $param_error );
    }

    if ( $global_sc_cache->hasEntry($dn) ) {
        my $result = $global_sc_cache->changeDeviceTypeForEntry(
            'dn'             => $dn,
            'device_type_dn' => $device_type_dn
        );
        if ( ref $result eq "HASH" ) {
            return ( 'changed_device' => $result );
        } else {
            return ( 'code' => 2,
                'message' => "Changing the device type failed" );
        }
    } else {
        $logger->warn("[LDAP] no entry found for dn: $dn");
        return ( 'code' => 1, 'message' => "No entry found for dn: $dn" );
    }
}

=head2 callSetUpdate( % )

    called via shared function queue

=over 4

=item Parameter

    struct

=item Return

    

=back

=cut

sub callSetUpdate
{
    my %params = (
        'location_dn'    => undef,
        'device_type_dn' => undef,
        'workstations'   => undef,
        'tftpboot'       => undef,
        'rsync'          => undef,
        @_,
    );

    my $logger = Log::Log4perl->get_logger();
    $logger->debug( Dumper(%params) );

    my $result = $global_sc_cache->setUpdate(
        'ldap' => $global_sc_ldap,
        %params
    );

    $logger->debug( "result: " . Dumper($result) );

    if ( !ref($result) ) {
        $logger->warn("unespected result $result");
        return;
    }

    # SharedFunction always expect a hash as return value
    return %{$result};
}

=head2 relocateDislocatedDevice( $ )

    called via shared function queue

=over 4

=item Parameter

    hash

=item Return

    see SmartClient::Leases2ldap->performAction

=back

=cut

sub relocateDislocatedDevice( $ )
{
    my ($href) = @_;

    my $logger = Log::Log4perl->get_logger();
    $logger->debug( Dumper($href) );

    my $id             = $href->{'id'};
    my $device_type_dn = $href->{'device_type_dn'};

    return $global_leases2ldap->performAction(
        id             => $id,
        permission     => PERM_GRANTED,
        device_type_dn => $device_type_dn
    );
}

=head2 notifyOfDelete

=over 4

=item Parameter

=item Return

=back

=cut

sub notifyOfDelete
{
    my ($aref_deleted_dn_list) = @_;

    my %result = ();
    my $logger = Log::Log4perl->get_logger();

    my @deleted_dn_list = @{ $aref_deleted_dn_list->[0] };

    $global_sc_cache->notifyDelete(@deleted_dn_list);

    foreach my $deleted_dn (@deleted_dn_list) {
        $logger->debug("got delete notification for $deleted_dn");
        $result{$deleted_dn} = 'OK';
    }
    return %result;
}

=head2 createConfig

=over 4

=item Parameter

=item Return

=back

=cut

sub createConfig
{
    my ($params) = @_;
    my %result = ();

    my $logger = Log::Log4perl->get_logger();

    $logger->debug("createConfig was called");
    $global_create_config_trigger = 1;
    $result{'state'} = 'OK';
    return %result;
}

=head2 callGetLeasesForDevice

=over 4

=item Parameter

=item Return

=back

=cut

sub callGetLeasesForDevice
{
    my ($mac_addr) = @_;
    
    my $href_result = $global_leases2ldap->getLeasesForMAC($mac_addr);
    return %{$href_result} if ( ref $href_result eq "HASH" );
    return;
}

=head2 callDeleteLeasesForDevice

=over 4

=item Parameter

=item Return

=back

=cut

sub callDeleteLeasesForDevice
{
    my ($mac_addr) = @_;
    
    my $href_result = $global_leases2ldap->triggerDeleteLeasesForMAC($mac_addr);
    return %{$href_result} if ( ref $href_result eq "HASH" );
    return;
}

# end of functions for SmartClient::SharedFunction
###########################################################

=head2 removeEmptyEntries

removes all empty strings from an array. Strings which consists only of spaces
will be considered as empty strings, too.

=over 4

=item Parameter

    array

=item Return

    array with empty strings removed

=back

=cut

sub removeEmptyEntries
{
    my @result = grep { !/^\s*$/ } @_;
    return @result;
}

=head2 readCurrentDhcpSubDirState

Reads the subdir name into which the next dhcp config files
should be generated.

The subdir name is relative to C<LDAP2DHCP_DHCPCONFPATH>.

=over 4

=item Parameters

reference to config

=item Return

name of the subdir

=back

=cut

sub readCurrentDhcpSubDirState
{
    my ($config) = @_;
    my $logger = Log::Log4perl->get_logger();

    my $current_subdir = 0;
    my $next_subdir = $config->get( 'LDAP2DHCP_DHCP_NEXT_SUBDIR' );
    if ( open( my $fh, '<', $next_subdir ) ) {
        while ( my $line = <$fh> ) {
            $current_subdir = int($line) % 2;
            last;
        }
        close($fh) or do {
            $logger->warn("Can not close file $next_subdir: $!");
        };
    } else {
        $logger->warn( 'Could not open file '
              . $next_subdir
              . " configured in LDAP2DHCP_DHCP_NEXT_SUBDIR to reconstruct current dhcp subdir: $!"
        );
    }
    return $current_subdir;
}

=head2 writeCurrentDhcpSubDirState

saves the name of the subdir into which the next dhcp config files
should be generated.

The subdir name is relative to C<LDAP2DHCP_DHCPCONFPATH>.

=over 4

=item Parameters

reference to config

name of the subdir

=item Return

=back

=cut

sub writeCurrentSubDirState
{
    my ( $config, $current_subdir ) = @_;
    my $logger = Log::Log4perl->get_logger();
    my $next_subdir = $config->get( 'LDAP2DHCP_DHCP_NEXT_SUBDIR' );
    if ( open( my $fh, '>', $next_subdir ) ) {
        print $fh $current_subdir;
        close($fh) or do {
            $logger->warn("Can not close $next_subdir: $!");
        };
    } else {
        $logger->warn( 'Could not write file '
              . $next_subdir
              . " from config LDAP2DHCP_DHCP_NEXT_SUBDIR: $!" );
    }
    return;
}

=head2 switchCurrentDhcpSubDir

Switch current subdir for ldap_generated dhcpd config files.

The subdir name is relative to C<LDAP2DHCP_DHCPCONFPATH>.

=over 4

=item Parameters

reference to config

current name of subdir for generated dhcpd config files

=item Return

name of next subdir where generated dhcpd config files should be written

=back

=cut

sub switchCurrentSubDir
{
    my ( $config, $current_dhcp_subdir ) = @_;
    my $logger = Log::Log4perl->get_logger();

    my $ldap_generated_basedir = $config->get( 'LDAP2DHCP_DHCPCONFPATH' );

    unlink("$ldap_generated_basedir/current");
    symlink( "$ldap_generated_basedir/$current_dhcp_subdir",
        "$ldap_generated_basedir/current" );
    my $next_configdir = $current_dhcp_subdir == 0 ? 1 : 0;
    copy( $_, "$ldap_generated_basedir/$next_configdir" )
      foreach (glob "$ldap_generated_basedir/$current_dhcp_subdir/*");
    $current_dhcp_subdir = $next_configdir;
    writeCurrentSubDirState( $config, $current_dhcp_subdir );
    $logger->info("Next dhcp_subdir will be $current_dhcp_subdir");

    return $current_dhcp_subdir;
}

###########################################################

=head2 main

=cut

mainLoop();

